import unittest
import odmlib.odm_1_3_2.model as ODM
import odmlib.odm_loader as OL
import odmlib.loader as LD
import os


class TestInsertItem(unittest.TestCase):
    def setUp(self):
        self.odm_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'cdash-odm-test.xml')
        self.odm_file_out = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'cdash-odm-test-insert.xml')
        self.loader = LD.ODMLoader(OL.XMLODMLoader())

    def test_insert_item_with_none_element(self):
        attrs = self.set_item_attributes()
        item = ODM.ItemDef(**attrs)
        # to support the test a null version of CodeListRef is created, but should be ignored on output
        if item.CodeListRef:
            self.assertIsNotNone(item.CodeListRef)
        if item.SignificantDigits:
            self.assertIsNotNone(item.SignificantDigits)
        self.loader.open_odm_document(self.odm_file)
        mdv = self.loader.MetaDataVersion()
        mdv.ItemDef.append(item)
        attrs = self.get_root_attributes()
        root = ODM.ODM(**attrs)
        study = self.add_study(mdv)
        root.Study = [study]
        root.write_xml(self.odm_file_out)
        loader = LD.ODMLoader(OL.XMLODMLoader())
        loader.open_odm_document(self.odm_file_out)
        mdv = loader.MetaDataVersion()
        is_found_inserted_item = False
        for it in mdv.ItemDef:
            if it.OID == "ODM.IT.AE.TEST":
                is_found_inserted_item = True
        self.assertTrue(is_found_inserted_item)
        # TODO test json for None values
        odm_json = root.to_json()

    def add_study(self, mdv):
        study_name = ODM.StudyName(_content="ODM XML Test Study Name")
        protocol_name = ODM.ProtocolName(_content="ODM XML Test Study")
        study_description = ODM.StudyDescription(_content="Testing the generation of an ODM XML file")
        gv = ODM.GlobalVariables()
        gv.StudyName = study_name
        gv.StudyDescription = study_description
        gv.ProtocolName = protocol_name
        study = ODM.Study(OID="ODM.STUDY.001")
        study.GlobalVariables = gv
        study.MetaDataVersion = [mdv]
        return study

    def set_item_attributes(self):
            return {"OID": "ODM.IT.AE.TEST", "Name": "Any AEs?", "DataType": "text", "Length": 1, "SASFieldName": "AEYN",
                "SDSVarName": "AEYN", "Origin": "CRF", "Comment": None}

    def get_root_attributes(self):
        return {"FileOID": "ODM.MDV.TEST.001", "Granularity": "Metadata",
                "AsOfDateTime": "2020-07-13T00:13:51.309617+00:00",
                "CreationDateTime": "2020-07-13T00:13:51.309617+00:00", "ODMVersion": "1.3.2", "FileType": "Snapshot",
                "Originator": "RDS", "SourceSystem": "ODMLib", "SourceSystemVersion": "0.1"}



import xml.etree.ElementTree as ET
import xmlschema as XSD
import odmlib.ns_registry as NS
from abc import ABC, abstractmethod
import json

ODM_NS = {'odm': 'http://www.cdisc.org/ns/odm/v1.3'}
ODM_PREFIX = "odm:"


class SchemaValidator(ABC):
    @abstractmethod
    def validate_tree(self, tree):
        raise NotImplementedError(
            "Attempted to execute an abstract method validate_tree in the Validator class")

    @abstractmethod
    def validate_file(self, xml_file):
        raise NotImplementedError(
            "Attempted to execute an abstract method validate_file in the Validator class")


class ODMSchemaValidator(SchemaValidator):
    def __init__(self, xsd_file):
        self.xsd = XSD.XMLSchema(xsd_file)

    def validate_tree(self, tree):
        result = self.xsd.is_valid(tree)
        return result

    def validate_file(self, odm_file):
        result = self.xsd.validate(odm_file)
        return result


class BaseParser:
    def __init__(self, ns_registry):
        if ns_registry:
            self.nsr = ns_registry
        else:
            self.nsr = NS.NamespaceRegistry(prefix="odm", uri="http://www.cdisc.org/ns/odm/v1.3", is_default=True)

    def register_namespaces(self):
        for prefix, url in self.nsr.namespaces.items():
            ET.register_namespace(prefix, url)

    def __getattr__(self, item):
        """ enables the parser to dynamically parse any element given it's parent """
        def parse_method(*args, parent, ns_prefix="odm", **kwargs):
            elem_list = []
            for elem in parent.findall(ns_prefix + ":" + item, self.nsr.get_ns_entry_dict(ns_prefix)):
                elem_list.append({**elem.attrib, "elem": elem})
            return elem_list
        return parse_method


class ElementParser:
    def __init__(self):
        self.root = None
        self.mdv = []
        self.admin_data = []
        self.clinical_data = []

    def ODM(self):
        return self.root

    def Study(self):
        study = self.root.find(ODM_PREFIX + "Study", ODM_NS)
        return study

    def MetaDataVersion(self):
        study = self.root.find(ODM_PREFIX + "Study", ODM_NS)
        self.mdv = study.findall(ODM_PREFIX + "MetaDataVersion", ODM_NS)
        return self.mdv

    def AdminData(self):
        self.admin_data = self.root.findall(ODM_PREFIX + "AdminData", ODM_NS)
        return self.admin_data

    def ClinicalData(self):
        self.clinical_data = self.root.findall(ODM_PREFIX + "ClinicalData", ODM_NS)
        return self.clinical_data

    def ReferenceData(self):
        self.reference_data = self.root.findall(ODM_PREFIX + "ReferenceData", ODM_NS)
        return self.reference_data


class ODMParser(BaseParser, ElementParser):
    def __init__(self, odm_file, namespace_registry=None):
        self.odm_file = odm_file
        super().__init__(ns_registry=namespace_registry)

    def parse(self):
        self.register_namespaces()
        odm_tree = ET.parse(self.odm_file)
        self.root = odm_tree.getroot()
        return self.root

    def parse_tree(self):
        self.register_namespaces()
        return ET.parse(self.odm_file)


class ODMStringParser(BaseParser, ElementParser):
    def __init__(self, odm_string, namespace_registry=None):
        self.odm_string = odm_string
        super().__init__(ns_registry=namespace_registry)

    def parse(self):
        self.register_namespaces()
        self.root = ET.fromstring(self.odm_string)
        return self.root

    def parse_tree(self):
        self.register_namespaces()
        return ET.fromstring(self.odm_string)


class ODMJSONStringParser:
    def __init__(self, odm_string):
        self.root = json.loads(odm_string)
        self.mdv = []
        self.admin_data = []
        self.clinical_data = []
        self.reference_data = []

    def parse(self):
        return self.root

    def ODM(self):
        return self.root

    def Study(self):
        study = self.root["Study"]
        return study

    def MetaDataVersion(self):
        study = self.root["Study"]
        self.mdv = study[0]["MetaDataVersion"]
        return self.mdv

    def AdminData(self):
        self.admin_data = self.root["AdminData"]
        return self.admin_data

    def ClinicalData(self):
        self.clinical_data = self.root["ClinicalData"]
        return self.clinical_data

    def ReferenceData(self):
        self.reference_data = self.root["ReferenceData"]
        return self.reference_data



import odmlib.document_loader as DL


class ODMLoader:
    """ loads an ODM-XML document into the object model """
    def __init__(self, odm_loader):
        if not isinstance(odm_loader, DL.DocumentLoader):
            raise TypeError("odm_loader argument must implement DocumentLoader")
        self.loader = odm_loader

    def create_odmlib(self, odm_doc, odm_key=None):
        odm_obj = self.loader.load_document(odm_doc, odm_key)
        return odm_obj

    def open_odm_document(self, filename):
        root = self.loader.create_document(filename)
        return root

    def load_odm_string(self, odm_string):
        root = self.loader.create_document_from_string(odm_string)
        return root

    def root(self):
        odm = self.loader.load_odm()
        return odm

    def MetaDataVersion(self, idx=0):
        mdv = self.loader.load_metadataversion(idx)
        return mdv

    def Study(self):
        study = self.loader.load_study()
        return study

    def __getattr__(self, attr):
        if hasattr(self.loader, attr):
            load_func = getattr(self.loader, attr, None)
            if callable(load_func):
                return load_func
            else:
                raise AttributeError(f"{attr} is not callable in the loader {self.loader.__class__.__name__}")
        else:
            raise AttributeError(f"The {attr} method does not exist in the loader {self.loader.__class__.__name__}")



from unittest import TestCase
import odmlib.odm_1_3_2.model as ODM
import odmlib.odm_1_3_2.rules.metadata_schema as METADATA
import odmlib.odm_1_3_2.rules.oid_ref as OID


class TestConformanceCheckODM(TestCase):
    def setUp(self):
        self.validator = METADATA.MetadataSchema()
        self.oid_checker = OID.OIDRef()

    def test_OID_unique_clean(self):
        attrs = {"OID": "MDV.TRACE-XML-ODM-01", "Name": "TRACE-XML MDV", "Description": "Trace-XML Example"}
        self.mdv = ODM.MetaDataVersion(**attrs)
        self.mdv.Protocol = self.add_protocol()
        self.mdv.StudyEventDef = self.add_SED()
        self.mdv.FormDef = self.add_FD()
        self.mdv.ItemGroupDef = self.add_IGD()
        self.mdv.ItemDef = self.add_ITD()
        self.mdv.CodeList = self.add_CL()
        self.mdv.MethodDef = self.add_MD()
        self.mdv.ConditionDef = self.add_CD()
        # oid uniqueness check and ref/def init
        self.mdv.verify_oids(self.oid_checker)
        # oid ref/def check
        self.assertTrue(self.oid_checker.check_oid_refs())

    def test_OID_unique_dirty(self):
        attrs = {"OID": "MDV.TRACE-XML-ODM-01", "Name": "TRACE-XML MDV", "Description": "Trace-XML Example"}
        self.mdv = ODM.MetaDataVersion(**attrs)
        self.mdv.Protocol = self.add_protocol()
        self.mdv.StudyEventDef = self.add_SED()
        self.mdv.FormDef = self.add_FD()
        self.mdv.ItemGroupDef = self.add_IGD()
        self.mdv.ItemDef = self.add_ITD_nonunique()
        self.mdv.CodeList = self.add_CL()
        self.mdv.MethodDef = self.add_MD()
        self.mdv.ConditionDef = self.add_CD()
        oid_checker = OID.OIDRef()
        with self.assertRaises(ValueError):
            # checks for non-unique OIDs and runs the ref/def check
            self.mdv.verify_oids(oid_checker)

    def test_OID_refdef_dirty(self):
        attrs = {"OID": "MDV.TRACE-XML-ODM-01", "Name": "TRACE-XML MDV", "Description": "Trace-XML Example"}
        self.mdv = ODM.MetaDataVersion(**attrs)
        self.mdv.Protocol = self.add_protocol()
        self.mdv.StudyEventDef = self.add_SED()
        self.mdv.FormDef = self.add_FD()
        self.mdv.ItemGroupDef = self.add_IGD_refdef()
        self.mdv.ItemDef = self.add_ITD()
        self.mdv.CodeList = self.add_CL()
        self.mdv.MethodDef = self.add_MD()
        self.mdv.ConditionDef = self.add_CD()
        oid_checker = OID.OIDRef()
        with self.assertRaises(ValueError):
            # checks for non-unique OIDs and runs the ref/def check
            self.mdv.verify_oids(oid_checker)

    def test_OID_creation_dirty(self):
        attrs = {"OID": "MDV.TRACE-XML-ODM-01", "Description": "Trace-XML Example"}
        with self.assertRaises(ValueError):
            # MDV creation should fail due to missing Name attribute
            self.mdv = ODM.MetaDataVersion(**attrs)

    def test_MethodDef_missing_Description(self):
        attrs = {"OID": "MDV.TRACE-XML-ODM-01", "Name": "TRACE-XML MDV", "Description": "Trace-XML Example"}
        self.mdv = ODM.MetaDataVersion(**attrs)
        tt1 = ODM.TranslatedText(_content="Age at Screening Date (Screening Date - Birth date)", lang="en")
        desc1 = ODM.Description()
        desc1.TranslatedText.append(tt1)
        # MethodDef creation that includes the Description object
        md1 = ODM.MethodDef(OID="ODM.MT.AGE", Name="Algorithm to dervive AGE", Type="Computation", Description=desc1)
        self.mdv.MethodDef.append(md1)
        # MethodDef creation succeeds without Description object
        md2 = ODM.MethodDef(OID="ODM.MT.DOB", Name="Create BRTHDTC from date ELEMENTS", Type="Computation")
        self.mdv.MethodDef.append(md2)
        # test that both MethodDefs were added to MDV
        self.assertEqual(len(self.mdv.MethodDef), 2)
        # test that the content of Description exists for the first MethodDef
        self.assertEqual(self.mdv.MethodDef[0].Description.TranslatedText[0]._content, "Age at Screening Date (Screening Date - Birth date)")
        validator = METADATA.MetadataSchema()
        # this is how you check conformance that includes required elements, not just attributes
        is_valid = self.mdv.MethodDef[0].verify_conformance(validator)
        # Test shows that MethodDef conforms to schema
        self.assertTrue(is_valid)
        print(self.mdv.MethodDef[0].to_dict())
        print(self.mdv.MethodDef[1].to_dict())
        with self.assertRaises(ValueError):
            # test shows that MethodDef two fails the schema check because it is missing the required Description object
            self.mdv.MethodDef[1].verify_conformance(validator)

    def add_protocol(self):
        p = ODM.Protocol()
        tt = ODM.TranslatedText(_content="Trace-XML Test CDASH File", lang="en")
        p.Description = ODM.Description()
        p.Description.TranslatedText = [tt]
        ser1 = ODM.StudyEventRef(StudyEventOID="BASELINE", OrderNumber=1, Mandatory="Yes")
        ser2 = ODM.StudyEventRef(StudyEventOID="FOLLOW-UP", OrderNumber=2, Mandatory="Yes")
        p.StudyEventRef = [ser1, ser2]
        p.Alias = [ODM.Alias(Context="ClinicalTrials.gov", Name="trace-protocol")]
        return p

    def add_SED(self):
        fr1 = ODM.FormRef(FormOID="ODM.F.DM", Mandatory="Yes", OrderNumber=1)
        fr2 = ODM.FormRef(FormOID="ODM.F.VS", Mandatory="Yes", OrderNumber=2)
        fr3 = ODM.FormRef(FormOID="ODM.F.AE", Mandatory="Yes", OrderNumber=3)
        ser1 = ODM.StudyEventDef(OID="BASELINE", Name="Baseline Visit", Repeating="No", Type="Scheduled")
        ser1.FormRef = [fr1, fr2, fr3]
        ser2 = ODM.StudyEventDef(OID="FOLLOW-UP", Name="Follow-up Visit", Repeating="Yes", Type="Scheduled")
        ser2.FormRef = [fr1, fr2, fr3]
        return [ser1, ser2]

    def add_FD(self):
        igr1 = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.Common", Mandatory="Yes", OrderNumber=1)
        #igr2 = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.VS_GENERAL", Mandatory="Yes", OrderNumber=2)
        igr3 = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.VS", Mandatory="Yes", OrderNumber=3)
        fd1 = ODM.FormDef(OID="ODM.F.VS", Name="Vital Signs", Repeating="No")
        #fd1.ItemGroupRef = [igr1, igr2, igr3]
        fd1.ItemGroupRef = [igr1, igr3]
        igr4 = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.Common", Mandatory="Yes", OrderNumber=1)
        igr5 = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.DM", Mandatory="Yes", OrderNumber=2)
        fd2 = ODM.FormDef(OID="ODM.F.DM", Name="Demographics", Repeating="No")
        fd2.ItemGroupRef = [igr4, igr5]
        igr6 = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.Common", Mandatory="Yes", OrderNumber=1)
        igr7 = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.AE", Mandatory="Yes", OrderNumber=2)
        fd3 = ODM.FormDef(OID="ODM.F.AE", Name="Adverse Events", Repeating="No")
        fd3.ItemGroupRef = [igr6, igr7]
        return [fd1, fd2, fd3]

    def add_IGD(self):
        itr1 = ODM.ItemRef(ItemOID="ODM.IT.VS.VSDAT", Mandatory="Yes")
        itr2 = ODM.ItemRef(ItemOID="ODM.IT.VS.BP.DIABP.VSORRES", Mandatory="Yes")
        itr3 = ODM.ItemRef(ItemOID="ODM.IT.VS.BP.SYSBP.VSORRES", Mandatory="Yes")
        igd1 = ODM.ItemGroupDef(OID="ODM.IG.VS", Name="Vital Sign Measurement", Repeating="Yes")
        igd1.ItemRef = [itr1, itr2, itr3]
        igr4 = ODM.ItemRef(ItemOID="ODM.IT.DM.BRTHYR", Mandatory="Yes")
        igr5 = ODM.ItemRef(ItemOID="ODM.IT.DM.SEX", Mandatory="Yes")
        igd2 = ODM.ItemGroupDef(OID="ODM.IG.DM", Name="Demographics", Repeating="No")
        igd2.ItemRef = [igr4, igr5]
        igr6 = ODM.ItemRef(ItemOID="ODM.IT.Common.SubjectID", Mandatory="Yes")
        igr7 = ODM.ItemRef(ItemOID="ODM.IT.Common.Visit", Mandatory="Yes")
        igd3 = ODM.ItemGroupDef(OID="ODM.IG.Common", Name="Common", Repeating="No")
        igd3.ItemRef = [igr6, igr7]
        igr8 = ODM.ItemRef(ItemOID="ODM.IT.AE.AETERM", Mandatory="Yes")
        igr9 = ODM.ItemRef(ItemOID="ODM.IT.AE.AESEV", Mandatory="Yes")
        igd4 = ODM.ItemGroupDef(OID="ODM.IG.AE", Name="Adverse Events", Repeating="Yes")
        igd4.ItemRef = [igr8, igr9]
        return [igd1, igd2, igd3, igd4]

    def add_IGD_refdef(self):
        itr1 = ODM.ItemRef(ItemOID="ODM.IT.VS.VSDAT", Mandatory="Yes")
        itr2 = ODM.ItemRef(ItemOID="ODM.IT.VS.BP.DIABP.VSORRES_BAD", Mandatory="Yes")
        itr3 = ODM.ItemRef(ItemOID="ODM.IT.VS.BP.SYSBP.VSORRES", Mandatory="Yes")
        igd1 = ODM.ItemGroupDef(OID="ODM.IG.VS", Name="Vital Sign Measurement", Repeating="Yes")
        igd1.ItemRef = [itr1, itr2, itr3]
        igr4 = ODM.ItemRef(ItemOID="ODM.IT.DM.BRTHYR", Mandatory="Yes")
        igr5 = ODM.ItemRef(ItemOID="ODM.IT.DM.SEX", Mandatory="Yes")
        igd2 = ODM.ItemGroupDef(OID="ODM.IG.DM", Name="Demographics", Repeating="No")
        igd2.ItemRef = [igr4, igr5]
        igr6 = ODM.ItemRef(ItemOID="ODM.IT.Common.SubjectID", Mandatory="Yes")
        igr7 = ODM.ItemRef(ItemOID="ODM.IT.Common.Visit", Mandatory="Yes")
        igd3 = ODM.ItemGroupDef(OID="ODM.IG.Common", Name="Common", Repeating="No")
        igd3.ItemRef = [igr6, igr7]
        igr8 = ODM.ItemRef(ItemOID="ODM.IT.AE.AETERM", Mandatory="Yes")
        igr9 = ODM.ItemRef(ItemOID="ODM.IT.AE.AESEV", Mandatory="Yes")
        igd4 = ODM.ItemGroupDef(OID="ODM.IG.AE", Name="Adverse Events", Repeating="Yes")
        igd4.ItemRef = [igr8, igr9]
        return [igd1, igd2, igd3, igd4]

    def add_ITD(self):
        # ItemDef 1
        ttd1 = ODM.TranslatedText(_content="Date of measurements", lang="en")
        ttq1 = ODM.TranslatedText(_content="Date", lang="en")
        desc1 = ODM.Description()
        desc1.TranslatedText = [ttd1]
        q1 = ODM.Question()
        q1.TranslatedText = [ttq1]
        a1 = ODM.Alias(Context="CDASH", Name="VSDAT")
        itd1 = ODM.ItemDef(OID="ODM.IT.VS.VSDAT", Name="Date", DataType="partialDate")
        itd1.Description = desc1
        itd1.Question = q1
        itd1.Alias = [a1]
        # ItemDef 2
        ttd2 = ODM.TranslatedText(_content="Result of the vital signs measurement as originally received or collected.", lang="en")
        ttq2 = ODM.TranslatedText(_content="Diastolic", lang="en")
        desc2 = ODM.Description()
        desc2.TranslatedText = [ttd2]
        q2 = ODM.Question()
        q2.TranslatedText = [ttq2]
        a2a = ODM.Alias(Context="CDASH", Name="BP.DIABP.VSORRES")
        a2b = ODM.Alias(Context="CDASH/SDTM", Name="VSORRES+VSORRESU")
        itd2 = ODM.ItemDef(OID="ODM.IT.VS.BP.VSORRESU", Name="BP Units", DataType="text")
        itd2.Description = desc2
        itd2.Question = q2
        itd2.Alias = [a2a, a2b]
        # ItemDef 3
        ttd3 = ODM.TranslatedText(_content="Adverse Event Term", lang="en")
        ttq3 = ODM.TranslatedText(_content="AE Term", lang="en")
        desc3 = ODM.Description()
        desc3.TranslatedText = [ttd3]
        q3 = ODM.Question()
        q3.TranslatedText = [ttq3]
        itd3 = ODM.ItemDef(OID="ODM.IT.AE.AETERM", Name="AE Term", DataType="text")
        itd3.Description = desc3
        itd3.Question = q3
        # ItemDef 4
        ttd4 = ODM.TranslatedText(_content="Adverse Event Severity", lang="en")
        ttq4 = ODM.TranslatedText(_content="AE Severity", lang="en")
        desc4 = ODM.Description()
        desc4.TranslatedText = [ttd4]
        q4 = ODM.Question()
        q4.TranslatedText = [ttq4]
        itd4 = ODM.ItemDef(OID="ODM.IT.AE.AESEV", Name="AE Severity", DataType="text")
        itd4.Description = desc4
        itd4.Question = q4
        # ItemDef 5
        ttd5 = ODM.TranslatedText(_content="Subject ID", lang="en")
        ttq5 = ODM.TranslatedText(_content="Subject ID", lang="en")
        desc5 = ODM.Description()
        desc5.TranslatedText = [ttd5]
        q5 = ODM.Question()
        q5.TranslatedText = [ttq5]
        itd5 = ODM.ItemDef(OID="ODM.IT.Common.SubjectID", Name="Subject ID", DataType="text")
        itd5.Description = desc5
        itd5.Question = q5
        # ItemDef 6
        ttd6 = ODM.TranslatedText(_content="Diastolic Blood Pressure Result", lang="en")
        ttq6 = ODM.TranslatedText(_content="Diastolic BP", lang="en")
        desc6 = ODM.Description()
        desc6.TranslatedText = [ttd6]
        q6 = ODM.Question()
        q6.TranslatedText = [ttq6]
        itd6 = ODM.ItemDef(OID="ODM.IT.VS.BP.DIABP.VSORRES", Name="DBP Result", DataType="text")
        itd6.Description = desc6
        itd6.Question = q6
        # ItemDef 7
        ttd7 = ODM.TranslatedText(_content="Birth Year", lang="en")
        ttq7 = ODM.TranslatedText(_content="DOB Year", lang="en")
        desc7 = ODM.Description()
        desc7.TranslatedText = [ttd7]
        q7 = ODM.Question()
        q7.TranslatedText = [ttq7]
        itd7 = ODM.ItemDef(OID="ODM.IT.DM.BRTHYR", Name="Birth Year", DataType="text")
        itd7.Description = desc7
        itd7.Question = q7
        # ItemDef 8
        ttd8 = ODM.TranslatedText(_content="Visit", lang="en")
        ttq8 = ODM.TranslatedText(_content="Visit", lang="en")
        desc8 = ODM.Description()
        desc8.TranslatedText = [ttd8]
        q8 = ODM.Question()
        q8.TranslatedText = [ttq8]
        itd8 = ODM.ItemDef(OID="ODM.IT.Common.Visit", Name="Visit", DataType="text")
        itd8.Description = desc8
        itd8.Question = q8
        # ItemDef 9
        ttd9 = ODM.TranslatedText(_content="Sex", lang="en")
        ttq9 = ODM.TranslatedText(_content="Sex", lang="en")
        desc9 = ODM.Description()
        desc9.TranslatedText = [ttd9]
        q9 = ODM.Question()
        q9.TranslatedText = [ttq9]
        itd9 = ODM.ItemDef(OID="ODM.IT.DM.SEX", Name="Sex", DataType="text")
        itd9.Description = desc9
        itd9.Question = q9
        # ItemDef 10
        ttd10 = ODM.TranslatedText(_content="Systolic Blood Pressure Result", lang="en")
        ttq10 = ODM.TranslatedText(_content="Systolic BP", lang="en")
        desc10 = ODM.Description()
        desc10.TranslatedText = [ttd10]
        q10 = ODM.Question()
        q10.TranslatedText = [ttq10]
        itd10 = ODM.ItemDef(OID="ODM.IT.VS.BP.SYSBP.VSORRES", Name="Systolic BP Result", DataType="text")
        itd10.Description = desc10
        itd10.Question = q10
        return [itd1, itd2, itd3, itd4, itd5, itd6, itd7, itd8, itd9, itd10]


    def add_ITD_nonunique(self):
        # ItemDef 1
        ttd1 = ODM.TranslatedText(_content="Date of measurements", lang="en")
        ttq1 = ODM.TranslatedText(_content="Date", lang="en")
        desc1 = ODM.Description()
        desc1.TranslatedText = [ttd1]
        q1 = ODM.Question()
        q1.TranslatedText = [ttq1]
        a1 = ODM.Alias(Context="CDASH", Name="VSDAT")
        itd1 = ODM.ItemDef(OID="ODM.IT.VS.VSDAT", Name="Date", DataType="partialDate")
        itd1.Description = desc1
        itd1.Question = q1
        itd1.Alias = [a1]
        # ItemDef 2
        ttd2 = ODM.TranslatedText(_content="Result of the vital signs measurement as originally received or collected.", lang="en")
        ttq2 = ODM.TranslatedText(_content="Diastolic", lang="en")
        desc2 = ODM.Description()
        desc2.TranslatedText = [ttd2]
        q2 = ODM.Question()
        q2.TranslatedText = [ttq2]
        a2a = ODM.Alias(Context="CDASH", Name="BP.DIABP.VSORRES")
        a2b = ODM.Alias(Context="CDASH/SDTM", Name="VSORRES+VSORRESU")
        itd2 = ODM.ItemDef(OID="ODM.IT.VS.BP.VSORRESU", Name="BP Units", DataType="text")
        itd2.Description = desc2
        itd2.Question = q2
        itd2.Alias = [a2a, a2b]
        # ItemDef 3
        ttd3 = ODM.TranslatedText(_content="Adverse Event Term", lang="en")
        ttq3 = ODM.TranslatedText(_content="AE Term", lang="en")
        desc3 = ODM.Description()
        desc3.TranslatedText = [ttd3]
        q3 = ODM.Question()
        q3.TranslatedText = [ttq3]
        itd3 = ODM.ItemDef(OID="ODM.IT.AE.AETERM", Name="AE Term", DataType="text")
        itd3.Description = desc3
        itd3.Question = q3
        # ItemDef 4
        ttd4 = ODM.TranslatedText(_content="Adverse Event Severity", lang="en")
        ttq4 = ODM.TranslatedText(_content="AE Severity", lang="en")
        desc4 = ODM.Description()
        desc4.TranslatedText = [ttd4]
        q4 = ODM.Question()
        q4.TranslatedText = [ttq4]
        itd4 = ODM.ItemDef(OID="ODM.IT.AE.AESEV", Name="AE Severity", DataType="text")
        itd4.Description = desc4
        itd4.Question = q4
        # ItemDef 5
        ttd5 = ODM.TranslatedText(_content="Subject ID", lang="en")
        ttq5 = ODM.TranslatedText(_content="Subject ID", lang="en")
        desc5 = ODM.Description()
        desc5.TranslatedText = [ttd5]
        q5 = ODM.Question()
        q5.TranslatedText = [ttq5]
        itd5 = ODM.ItemDef(OID="ODM.IT.Common.SubjectID", Name="Subject ID", DataType="text")
        itd5.Description = desc5
        itd5.Question = q5
        # ItemDef 6
        ttd6 = ODM.TranslatedText(_content="Diastolic Blood Pressure Result", lang="en")
        ttq6 = ODM.TranslatedText(_content="Diastolic BP", lang="en")
        desc6 = ODM.Description()
        desc6.TranslatedText = [ttd6]
        q6 = ODM.Question()
        q6.TranslatedText = [ttq6]
        itd6 = ODM.ItemDef(OID="ODM.IT.VS.BP.SYSBP.VSORRES", Name="DBP Result", DataType="text")
        itd6.Description = desc6
        itd6.Question = q6
        # ItemDef 7
        ttd7 = ODM.TranslatedText(_content="Birth Year", lang="en")
        ttq7 = ODM.TranslatedText(_content="DOB Year", lang="en")
        desc7 = ODM.Description()
        desc7.TranslatedText = [ttd7]
        q7 = ODM.Question()
        q7.TranslatedText = [ttq7]
        itd7 = ODM.ItemDef(OID="ODM.IT.DM.BRTHYR", Name="Birth Year", DataType="text")
        itd7.Description = desc7
        itd7.Question = q7
        # ItemDef 8
        ttd8 = ODM.TranslatedText(_content="Visit", lang="en")
        ttq8 = ODM.TranslatedText(_content="Visit", lang="en")
        desc8 = ODM.Description()
        desc8.TranslatedText = [ttd8]
        q8 = ODM.Question()
        q8.TranslatedText = [ttq8]
        itd8 = ODM.ItemDef(OID="ODM.IT.Common.Visit", Name="Visit", DataType="text")
        itd8.Description = desc8
        itd8.Question = q8
        # ItemDef 9
        ttd9 = ODM.TranslatedText(_content="Sex", lang="en")
        ttq9 = ODM.TranslatedText(_content="Sex", lang="en")
        desc9 = ODM.Description()
        desc9.TranslatedText = [ttd9]
        q9 = ODM.Question()
        q9.TranslatedText = [ttq9]
        itd9 = ODM.ItemDef(OID="ODM.IT.DM.SEX", Name="Sex", DataType="text")
        itd9.Description = desc9
        itd9.Question = q9
        # ItemDef 10
        ttd10 = ODM.TranslatedText(_content="Systolic Blood Pressure Result", lang="en")
        ttq10 = ODM.TranslatedText(_content="Systolic BP", lang="en")
        desc10 = ODM.Description()
        desc10.TranslatedText = [ttd10]
        q10 = ODM.Question()
        q10.TranslatedText = [ttq10]
        itd10 = ODM.ItemDef(OID="ODM.IT.VS.BP.SYSBP.VSORRES", Name="Systolic BP Result", DataType="text")
        itd10.Description = desc10
        itd10.Question = q10
        return [itd1, itd2, itd3, itd4, itd5, itd6, itd7, itd8, itd9, itd10]

    def add_CD(self):
        tt1 = ODM.TranslatedText(_content="Skip the BRTHMO field when BRTHYR length NE 4", lang="en")
        desc = ODM.Description()
        desc.TranslatedText = [tt1]
        cd = ODM.ConditionDef(OID="ODM.CD.BRTHMO", Name="Skip BRTHMO when no BRTHYR")
        cd.Description = desc
        return [cd]

    def add_MD(self):
        tt1 = ODM.TranslatedText(_content="Concatenation of BRTHYR, BRTHMO, and BRTHDY in ISO 8601 format", lang="en")
        desc = ODM.Description()
        desc.TranslatedText = [tt1]
        md = ODM.MethodDef(OID="ODM.MT.DOB", Name="Create BRTHDTC from date ELEMENTS", Type="Computation")
        md.Description = desc
        return [md]

    def add_CL(self):
        tt1 = ODM.TranslatedText(_content="No", lang="en")
        dc1 = ODM.Decode()
        dc1.TranslatedText = [tt1]
        cli1 = ODM.CodeListItem(CodedValue="N")
        cli1.Decode = dc1
        tt2 = ODM.TranslatedText(_content="Yes", lang="en")
        dc2 = ODM.Decode()
        dc2.TranslatedText = [tt2]
        cli2 = ODM.CodeListItem(CodedValue="Y")
        cli2.Decode = dc2
        cl = ODM.CodeList(OID="ODM.CL.NY_SUB_Y_N", Name="No Yes Response", DataType="text")
        cl.CodeListItem = [cli1, cli2]
        return [cl]





import odmlib.odm_element as OE
import odmlib.typed as T
import odmlib.ns_registry as NS


NS.NamespaceRegistry(prefix="odm", uri="http://www.cdisc.org/ns/odm/v1.3", is_default=True)
NS.NamespaceRegistry(prefix="nciodm", uri="http://ncicb.nci.nih.gov/xml/odm/EVS/CDISC")
NS.NamespaceRegistry(prefix="xs", uri="http://www.w3.org/2001/XMLSchema-instance")
NS.NamespaceRegistry(prefix="xml", uri="http://www.w3.org/XML/1998/namespace")
NS.NamespaceRegistry(prefix="xlink", uri="http://www.w3.org/1999/xlink")


class TranslatedText(OE.ODMElement):
    lang = T.String(namespace="xml")
    _content = T.String(required=True)


class Alias(OE.ODMElement):
    Context = T.String(required=True)
    Name = T.String(required=True)


class StudyDescription(OE.ODMElement):
    _content = T.String(required=True)


class ProtocolName(OE.ODMElement):
    _content = T.String(required=True)


class StudyName(OE.ODMElement):
    _content = T.String(required=True)


class GlobalVariables(OE.ODMElement):
    StudyName = T.ODMObject(element_class=StudyName)
    StudyDescription = T.ODMObject(element_class=StudyDescription)
    ProtocolName = T.ODMObject(element_class=ProtocolName)


class Description(OE.ODMElement):
    TranslatedText = T.ODMListObject(required=True, element_class=TranslatedText)


class CDISCSynonym(OE.ODMElement):
    _content = T.String(required=True)


class CDISCDefinition(OE.ODMElement):
    _content = T.String(required=True)


class CDISCSubmissionValue(OE.ODMElement):
    _content = T.String(required=True)


class PreferredTerm(OE.ODMElement):
    _content = T.String(required=True)


class EnumeratedItem(OE.ODMElement):
    CodedValue = T.String(required=True)
    ExtCodeID = T.String(required=True, namespace="nciodm")
    CDISCSynonym = T.ODMListObject(element_class=CDISCSynonym, namespace="nciodm")
    CDISCDefinition = T.ODMObject(required=True, element_class=CDISCDefinition, namespace="nciodm")
    PreferredTerm = T.ODMObject(required=True, element_class=PreferredTerm, namespace="nciodm")


class CodeList(OE.ODMElement):
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    DataType = T.ValueSetString(required=True)
    ExtCodeID = T.String(required=True, namespace="nciodm")
    CodeListExtensible = T.ValueSetString(required=True, namespace="nciodm")
    Description = T.ODMObject(element_class=Description)
    EnumeratedItem = T.ODMListObject(element_class=EnumeratedItem)
    CDISCSubmissionValue = T.ODMObject(required=True, element_class=CDISCSubmissionValue, namespace="nciodm")
    CDISCSynonym = T.ODMObject(required=True, element_class=CDISCSynonym, namespace="nciodm")
    PreferredTerm = T.ODMObject(required=True, element_class=PreferredTerm, namespace="nciodm")


class MetaDataVersion(OE.ODMElement):
    OID = T.OID(required=True)
    Name = T.Name(required=True)
    Description = T.String(required=False)
    CodeList = T.ODMListObject(element_class=CodeList)


class Study(OE.ODMElement):
    OID = T.String(required=True)
    GlobalVariables = T.ODMObject(required=True, element_class=GlobalVariables)
    MetaDataVersion = T.ODMListObject(required=False, element_class=MetaDataVersion)


class ODM(OE.ODMElement):
    Description = T.String(required=False)
    FileType = T.ValueSetString(required=True)
    Granularity = T.ValueSetString(required=False)
    FileOID = T.OID(required=True)
    CreationDateTime = T.DateTimeString(required=True)
    AsOfDateTime = T.DateTimeString(required=False)
    ODMVersion = T.ValueSetString(required=False)
    Originator = T.String(required=False)
    SourceSystem = T.String(required=False)
    SourceSystemVersion = T.String(required=False)
    schemaLocation = T.String(required=False, namespace="xs")
    Study = T.ODMListObject(element_class=Study)



from unittest import TestCase
import os
import odmlib.odm_parser as P
import xmlschema as XSD


class TestODMValidator(TestCase):
    def setUp(self) -> None:
        self.odm_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'cdash-odm-test.xml')
        # set the file and path to point to the odm 1.3.2 schema on your system
        odm_schema_file = os.path.join(os.sep, 'home', 'sam', 'standards', 'odm1-3-2', 'ODM1-3-2.xsd')
        self.validator = P.ODMSchemaValidator(odm_schema_file)

    def test_validate_tree_valid(self):
        self.parser = P.ODMParser(self.odm_file)
        tree = self.parser.parse_tree()
        is_valid = self.validator.validate_tree(tree)
        self.assertTrue(is_valid)

    def test_validate_file(self):
        # validate file raises an exception if validation fails
        self.assertIsNone(self.validator.validate_file(self.odm_file))

    def test_validate_file_invalid(self):
        odm_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'cdash-odm-test-invalid.xml')
        with self.assertRaises(XSD.validators.exceptions.XMLSchemaChildrenValidationError):
            self.validator.validate_file(odm_file)

    def test_validate_tree_invalid(self):
        odm_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'cdash-odm-test-invalid.xml')
        self.parser = P.ODMParser(odm_file)
        tree = self.parser.parse_tree()
        is_valid = self.validator.validate_tree(tree)
        self.assertFalse(is_valid)

    def test_validate_string_tree_valid(self):
        with open(self.odm_file, "r", encoding="utf-8") as f:
            self.odm_string = f.read()
        self.parser = P.ODMStringParser(self.odm_string)
        tree = self.parser.parse_tree()
        is_valid = self.validator.validate_tree(tree)
        self.assertTrue(is_valid)



from unittest import TestCase
import json
import odmlib.odm_1_3_2.model as ODM
import odmlib.odm_1_3_2.rules.metadata_schema as METADATA


class TestMethodDefCheck(TestCase):
    def setUp(self) -> None:
        attrs = self.set_attributes()
        self.methoddef = ODM.MethodDef(**attrs)

    def test_no_description(self):
        tt = ODM.TranslatedText(lang="en", _content="hello, world")
        validator = METADATA.MetadataSchema()
        results = tt.verify_conformance(validator)

        self.methoddef.FormalExpression = [ODM.FormalExpression(Context="Python 3.7", _content="print('hello world')")]
        self.assertEqual(self.methoddef.FormalExpression[0]._content, "print('hello world')")
        with self.assertRaises(ValueError):
            results = self.methoddef.verify_conformance(validator)

    def test_add_description(self):
        tt1 = ODM.TranslatedText(_content="Age at Screening Date (Screening Date - Birth date)", lang="en")
        tt2 = ODM.TranslatedText(_content="For the complete algorithm see the referenced external document.", lang="fr")
        self.methoddef.Description = ODM.Description()
        self.methoddef.Description.TranslatedText = [tt1, tt2]
        self.assertEqual(len(self.methoddef.Description.TranslatedText), 2)
        self.assertEqual(self.methoddef.Description.TranslatedText[1]._content,
                         'For the complete algorithm see the referenced external document.')

    def test_add_alias(self):
        self.methoddef.Alias = [ODM.Alias(Context="CDASH", Name="AGE")]
        self.assertEqual(self.methoddef.Alias[0].Name, "AGE")

    def test_to_json(self):
        attrs = self.set_attributes()
        method = ODM.MethodDef(**attrs)
        tt1 = ODM.TranslatedText(_content="Age at Screening Date (Screening Date - Birth date)", lang="en")
        method.Description = ODM.Description()
        method.Description.TranslatedText = [tt1]
        fex = ODM.FormalExpression(Context="Python 3.7", _content="print('hello world')")
        method.FormalExpression = [fex]
        method_dict = method.to_dict()
        print(method_dict)
        method_json = method.to_json()
        method_dict = json.loads(method_json)
        self.assertEqual(method_dict["OID"], "ODM.MT.AGE")
        self.assertDictEqual(method_dict, self.expected_dict())

    def test_to_dict(self):
        attrs = self.set_attributes()
        method = ODM.MethodDef(**attrs)
        desc = ODM.Description()
        tt1 = ODM.TranslatedText(_content="Age at Screening Date (Screening Date - Birth date)", lang="en")
        desc.TranslatedText = [tt1]
        method.Description = desc
        fex = ODM.FormalExpression(Context="Python 3.7", _content="print('hello world')")
        method.FormalExpression = [fex]
        method_dict = method.to_dict()
        self.assertEqual(method_dict["OID"], "ODM.MT.AGE")
        self.assertDictEqual(method_dict, self.expected_dict())

    def test_to_xml(self):
        attrs = self.set_attributes()
        method = ODM.MethodDef(**attrs)
        desc = ODM.Description()
        tt1 = ODM.TranslatedText(_content="Age at Screening Date (Screening Date - Birth date)", lang="en")
        desc.TranslatedText = [tt1]
        method.Description = desc
        fex = ODM.FormalExpression(Context="Python 3.7", _content="print('hello world')")
        method.FormalExpression = [fex]
        method_xml = method.to_xml()
        self.assertEqual(method_xml.attrib["OID"], "ODM.MT.AGE")
        fex = method_xml.find("FormalExpression")
        self.assertEqual(fex.attrib["Context"], "Python 3.7")


    def set_attributes(self):
        """
        set some MethodDef element attributes using test data
        :return: dictionary with MethodDef attribute information
        """
        return {"OID": "ODM.MT.AGE", "Name": "Algorithm to derive AGE", "Type": "Computation"}

    def expected_dict(self):
        return {'OID': 'ODM.MT.AGE', 'Name': 'Algorithm to derive AGE', 'Type': 'Computation', 'Description':
            {'TranslatedText': [{'_content': 'Age at Screening Date (Screening Date - Birth date)', 'lang': 'en'}]},
             'FormalExpression': [{'Context': 'Python 3.7', '_content': "print('hello world')"}]}



from setuptools import setup
import pathlib

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name='odmlib',
    version='0.1.3',
    packages=['tests', 'odmlib', 'odmlib.odm_1_3_2', 'odmlib.odm_1_3_2.rules', 'odmlib.define_2_0',
              'odmlib.define_2_0.rules', 'odmlib.define_2_1', 'odmlib.define_2_1.rules', 'odmlib.ct_1_1_1'],
    url='https://github.com/swhume/odmlib',
    license='MIT',
    author='Sam Hume',
    author_email='swhume@gmail.com',
    description='Work with ODM as Python objects',
    long_description=README,
    long_description_content_type="text/markdown",
    include_package_data=True,
    install_requires=[
        "xmlschema>=1.3.1", 
        "validators>=0.18.1", 
        "Cerberus>=1.3.2", 
        "pathvalidate>=2.3.1"
    ]
)



import odmlib.descriptor as DESC
import re
import odmlib.valueset as VS
import datetime
from validators import email as valid_email, url as valid_url
from pathvalidate import is_valid_filename


class Typed(DESC.Descriptor):
    odm_type = object

    def __set__(self, instance, value):
        if (value is not None) and not isinstance(value, self.odm_type):
            raise TypeError(f"Expected type {str(self.odm_type)} for {self.name} with value {str(value)}")
        # often does not call parent, calls next on __mro__ which maybe influenced by multiple inheritance
        super().__set__(instance, value)


class Integer(DESC.Descriptor):
    def __set__(self, instance, value):
        # in XML integers are presented as strings e.g. OrderNumber="1"
        if isinstance(value, str):
            try:
                value = int(value)
            except ValueError:
                raise TypeError(f"Expected string to convert to integer for {self.name} with a value {str(value)}")
        odm_type = int
        if (value is not None) and not isinstance(value, odm_type):
            raise TypeError(f"Expected type {str(odm_type)} for {self.name} with value {str(value)}")
        super().__set__(instance, value)


class Float(DESC.Descriptor):
    def __set__(self, instance, value):
        # in XML floats are presented as strings
        if isinstance(value, str):
            try:
                value = float(value)
            except ValueError:
                raise TypeError(f"Expected string to convert to float for {self.name} with a value {str(value)}")
        # convert integer into a float (e.g. 1 to 1.0)
        elif isinstance(value, int):
            value = float(value)
        if (value is not None) and not isinstance(value, float):
            raise TypeError(f"Expected type float for {self.name} with value {str(value)}")
        super().__set__(instance, value)


class String(Typed):
    odm_type = str


class OID(Typed):
    odm_type = str


class OIDRef(Typed):
    odm_type = str


class Name(Typed):
    odm_type = str


class ID(Typed):
    odm_type = str


class IDRef(Typed):
    odm_type = str


class List(Typed):
    odm_type = list


class Dictionary(Typed):
    odm_type = dict


class Positive(DESC.Descriptor):
    def __set__(self, instance, value):
        if (value is not None) and value <= 0:
            raise TypeError(f"Expected value > 0 for type Positive for {self.name}")
        super().__set__(instance, value)


class NonNegative(DESC.Descriptor):
    def __set__(self, instance, value):
        if (value is not None) and value < 0:
            raise TypeError(f"Expected value >= 0 for type Non-negative for {self.name}")
        super().__set__(instance, value)


class PositiveInteger(Integer, Positive):
    pass


class NonNegativeInteger(Integer, NonNegative):
    pass


class PositiveFloat(Float, Positive):
    pass


class NonNegativeFloat(Float, NonNegative):
    pass


class Sized(DESC.Descriptor):
    def __init__(self, *args, max_length, **kwargs):
        # pulls out named arg max_length and passes the remaining arguments along
        self.max_length = max_length
        super().__init__(*args, **kwargs)

    def __set__(self, instance, value):
        if (value is not None) and (len(value) > self.max_length):
            raise ValueError(f"{self.name} has a length of {len(value)} and exceeds the maximum length of "
                             f"{self.max_length}")
        super().__set__(instance, value)


class SizedString(String, Sized):
    pass


class Regex(DESC.Descriptor):
    """ pattern matching """
    def __init__(self, *args, pat, **kwargs):
        # takes a pattern arg named pat and passes the remaining arguments along
        self.pat = re.compile(pat)
        super().__init__(*args, **kwargs)

    def __set__(self, instance, value):
        if (value is not None) and not self.pat.match(value):
            raise ValueError(f"{self.name} has an invalid string of {value}")
        super().__set__(instance, value)


class SizedRegexString(SizedString, Regex):
    pass


class DateTimeString(DESC.Descriptor):
    def __set__(self, instance, value):
        iso_pat = r'^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$'
        pat = re.compile(iso_pat)
        if (value is not None) and (not pat.match(value)):
            raise ValueError(f"Expected type datetime for {self.name}, found value {value}")
        super().__set__(instance, value)


class PartialDateTimeString(DESC.Descriptor):
    def __set__(self, instance, value):
        part_pat = r'^((([0-9][0-9][0-9][0-9])((-(([0][1-9])|([1][0-2])))((-(([0][1-9])|([1-2][0-9])|([3][0-1])))(T((([0-1][0-9])|([2][0-3]))((:([0-5][0-9]))(((:([0-5][0-9]))((\.[0-9]+)?))?)?)?((((\+|-)(([0-1][0-9])|([2][0-3])):[0-5][0-9])|(Z)))?))?)?)?))$'
        compiled_part_pat = re.compile(part_pat)
        if (value is not None) and (not compiled_part_pat.match(value)):
            raise ValueError(f"Expected type PartialDateTime for {self.name}, found value {value}")
        super().__set__(instance, value)


class PartialDateString(DESC.Descriptor):
    def __set__(self, instance, value):
        if value and value.count('-') == 2:
            try:
                datetime.datetime.strptime(value, '%Y-%m-%d')
            except ValueError:
                raise ValueError(f"Expected type PartialDate for {self.name}, found value {value}")
        else:
            part_pat = r'^(([0-9][0-9][0-9][0-9])(-(([0][1-9])|([1][0-2])))?)$'
            compiled_part_pat = re.compile(part_pat)
            if (value is not None) and (not compiled_part_pat.match(value)):
                raise ValueError(f"Expected type PartialDate for {self.name}, found value {value}")
        super().__set__(instance, value)


class PartialTimeString(DESC.Descriptor):
    def __set__(self, instance, value):
        time_pat = r'^(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$'
        compiled_time_pat = re.compile(time_pat)
        part_pat = r'^((([0-1][0-9])|([2][0-3]))(:[0-5][0-9])?(((\+|-)(([0-1][0-9])|([2][0-3])):[0-5][0-9])|(Z))?)$'
        compiled_part_pat = re.compile(part_pat)
        if (value is not None) and (not (compiled_time_pat.match(value) or compiled_part_pat.match(value))):
            raise ValueError(f"Expected type IncompleteTime for {self.name}, found value {value}")
        super().__set__(instance, value)


class IncompleteDateTimeString(DESC.Descriptor):
    def __set__(self, instance, value):
        inc_pat = r'^(((([0-9][0-9][0-9][0-9]))|-)-(((([0][1-9])|([1][0-2])))|-)-(((([0][1-9])|([1-2][0-9])|([3][0-1])))|-)T(((([0-1][0-9])|([2][0-3])))|-):((([0-5][0-9]))|-):((([0-5][0-9](\.[0-9]+)?))|-)((((\+|-)(([0-1][0-9])|([2][0-3])):[0-5][0-9])|Z|-))?)$'
        compiled_inc_pat = re.compile(inc_pat)
        if (value is not None) and (not compiled_inc_pat.match(value)):
            raise ValueError(f"Expected type IncompleteDateTime for {self.name}, found value {value}")
        super().__set__(instance, value)


class IncompleteDateString(DESC.Descriptor):
    def __set__(self, instance, value):
        inc_pat = r'^(((([0-9][0-9][0-9][0-9]))|-)-(((([0][1-9])|([1][0-2])))|-)-(((([0][1-9])|([1-2][0-9])|([3][0-1])))|-))$'
        compiled_inc_pat = re.compile(inc_pat)
        if (value is not None) and (not compiled_inc_pat.match(value)):
            raise ValueError(f"Expected type IncompleteDate for {self.name}, found value {value}")
        super().__set__(instance, value)


class IncompleteTimeString(DESC.Descriptor):
    def __set__(self, instance, value):
        inc_pat = r'^((((([0-1][0-9])|([2][0-3])))|-):((([0-5][0-9]))|-):((([0-5][0-9](\.[0-9]+)?))|-)((((\+|-)(([0-1][0-9])|([2][0-3])):[0-5][0-9])|Z|-))?)$'
        compiled_inc_pat = re.compile(inc_pat)
        if (value is not None) and (not compiled_inc_pat.match(value)):
            raise ValueError(f"Expected type IncompleteTime for {self.name}, found value {value}")
        super().__set__(instance, value)


class DurationDateTimeString(DESC.Descriptor):
    def __set__(self, instance, value):
        dur_pat = r'^((\+ | -)?P([0-9]([0-9]+)?)W)$'
        pat = re.compile(dur_pat)
        if (value is not None) and (not pat.match(value)):
            raise ValueError(f"Expected type DurationDateTime for {self.name}, found value {value}")
        super().__set__(instance, value)


class DateString(DESC.Descriptor):
    def __set__(self, instance, value):
        try:
            if value is not None:
                datetime.datetime.strptime(value, '%Y-%m-%d')
        except ValueError:
            raise ValueError(f"Expected type date (YYYY-MM-DD) for {self.name}, found value {value}")
        super().__set__(instance, value)


class SASName(DESC.Descriptor):
    def __set__(self, instance, value):
        pat = re.compile("[A-Za-z_][A-Za-z0-9_]*$")
        if (value is not None) and (not pat.match(value) or len(value) > 8):
            raise ValueError(f"{self.name} has an invalid sasName of {value}")
        super().__set__(instance, value)


class SASFormat(DESC.Descriptor):
    def __set__(self, instance, value):
        # is_valid = True
        pat = re.compile("[A-Za-z_$][A-Za-z0-9_.]*$")
        if (value is not None) and (not pat.match(value) or len(value) > 8):
            raise ValueError(f"{self.name} has an invalid sasFormat of {value}")
        super().__set__(instance, value)


class Email(DESC.Descriptor):
    def __set__(self, instance, value):
        if (value is not None) and (not valid_email(value)):
            raise ValueError(f"{self.name} has an invalid email format of {value}")
        super().__set__(instance, value)


class Url(DESC.Descriptor):
    def __set__(self, instance, value):
        if (value is not None) and (not valid_url(value)):
            raise ValueError(f"{self.name} has an invalid url format of {value}")
        super().__set__(instance, value)


class FileName(DESC.Descriptor):
    def __set__(self, instance, value):
        if (value is not None) and (not is_valid_filename(value)):
            raise ValueError(f"{self.name} has an invalid fileName of {value}")
        super().__set__(instance, value)


class ValidValues(DESC.Descriptor):
    def __set__(self, instance, value):
        if (value is not None) and value not in VS.ValueSet.value_set(type(instance).__name__ + "." + self.name):
            raise TypeError(f"Invalid value {value} for {self.name}. Value must be one of "
                            f"{', '.join(VS.ValueSet.value_set(self.name))}")
        super().__set__(instance, value)


class ExtendedValidValues(DESC.Descriptor):
    def __set__(self, instance, value):
        if (value is not None) and value not in self.valid_values:
            raise TypeError(f"Invalid value {value} for {self.name}. Value must be one of "
                            f"{', '.join(self.valid_values)}")
        super().__set__(instance, value)


class ValueSetString(ValidValues, String):
    pass


class ODMObject(DESC.Descriptor):
    def __init__(self, *args, element_class,  **kwargs):
        self.obj_type = element_class
        kwargs["element_class"] = element_class
        self.namespace = kwargs.get("namespace", "")
        super().__init__(*args, **kwargs)

    def __set__(self, instance, value):
        if not isinstance(value, self.obj_type) and not isinstance(value, list):
            raise TypeError(f"The {self.name} object must be of type {self.obj_type}")
        super().__set__(instance, value)


class ODMListObject(ODMObject, list):
    def __set__(self, instance, value):
        if isinstance(value, list):
            for obj in value:
                if not isinstance(obj, self.obj_type):
                    raise TypeError(f"Every {self.name} object in the list must be of type {self.obj_type}")
        else:
            raise TypeError(f"The {self.name} object must be a list")
        super().__set__(instance, value)



#from collections import namedtuple

class OIDIndex:
    def __init__(self):
        self.oid_index = {}

    def add_oid(self, oid, element):
        """ odmlib expects all OIDs to be unique within the scope of an ODM document """
        if oid not in self.oid_index:
            self.oid_index[oid] = []
        self.oid_index[oid].append(element)

    def find_all(self, oid):
        if not self.oid_index:
            raise ValueError(f"The OID index is empty. Build the index prior to using find_all. OID {oid} not found.")
        elif oid not in self.oid_index:
            raise ValueError(f"OID {oid} not found in the OID index.")
        return self.oid_index[oid]



from unittest import TestCase
import json
import odmlib.odm_1_3_2.model as ODM


class TestFormDef(TestCase):
    def setUp(self) -> None:
        attrs = self.set_formdef_attributes()
        self.formdef = ODM.FormDef(**attrs)

    def test_add_description(self):
        tt1 = ODM.TranslatedText(_content="this is the first test description", lang="en")
        tt2 = ODM.TranslatedText(_content="this is the second test description", lang="en")
        self.formdef.Description = ODM.Description()
        self.formdef.Description.TranslatedText = [tt1, tt2]
        self.assertEqual(len(self.formdef.Description.TranslatedText), 2)
        self.assertEqual(self.formdef.Description.TranslatedText[1]._content, 'this is the second test description')

    def test_add_item_group_ref(self):
        igr = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.COMMON", Mandatory="Yes", OrderNumber=1)
        self.formdef.ItemGroupRef.append(igr)
        igr = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.VS_GENERAL", Mandatory="Yes", OrderNumber=2)
        self.formdef.ItemGroupRef.append(igr)
        igr = ODM.ItemGroupRef(ItemGroupOID="ODM.IG.VS", Mandatory="Yes", OrderNumber=3)
        self.formdef.ItemGroupRef.append(igr)
        self.assertEqual(self.formdef.ItemGroupRef[0].ItemGroupOID, "ODM.IG.COMMON")
        self.assertEqual(self.formdef.ItemGroupRef[2].OrderNumber, 3)

    def test_append_item_group_ref(self):
        fd = ODM.FormDef(OID="ODM.F.VS", Name="Vital Signs Form", Repeating="Yes")
        fd.Description = ODM.Description()
        fd.Description.TranslatedText.append(ODM.TranslatedText(_content="this is the first test description", lang="en"))
        fd.ItemGroupRef.append(ODM.ItemGroupRef(ItemGroupOID="ODM.IG.COMMON", Mandatory="Yes", OrderNumber=1))
        fd.ItemGroupRef.append(ODM.ItemGroupRef(ItemGroupOID="ODM.IG.VS_GENERAL", Mandatory="Yes", OrderNumber=2))
        fd.ItemGroupRef.append(ODM.ItemGroupRef(ItemGroupOID="ODM.IG.VS", Mandatory="Yes", OrderNumber=3))
        self.assertEqual(fd.ItemGroupRef[0].ItemGroupOID, "ODM.IG.COMMON")
        self.assertEqual(fd.ItemGroupRef[2].OrderNumber, 3)
        self.assertEqual(fd.Description.TranslatedText[0]._content, "this is the first test description")

    def test_add_alias(self):
        self.formdef.Alias.append(ODM.Alias(Context="SDTMIG", Name="VS"))
        self.formdef.Alias.append(ODM.Alias(Context="CDASHIG", Name="VS"))
        self.assertEqual(len(self.formdef.Alias), 2)
        self.assertEqual(self.formdef.Alias[1].Context, "CDASHIG")

    def test_add_not_alias(self):
        item = ODM.ItemDef(OID="ODM.IT.VSPOS", Name="VS Position", DataType="text")
        with self.assertRaises(TypeError):
            self.formdef.Alias = [item]
        self.formdef.Alias.append(ODM.Alias(Context="SDTMIG", Name="VS"))
        # list accepts invalid objects
        self.formdef.Alias.append(ODM.ItemDef(OID="ODM.IT.VSDT", Name="VS Date", DataType="text"))
        self.assertEqual(len(self.formdef.Alias), 2)
        self.assertEqual(self.formdef.Alias[0].Context, "SDTMIG")

    def test_to_json(self):
        attrs = self.set_formdef_attributes()
        fd = ODM.FormDef(**attrs)
        tt = ODM.TranslatedText(_content="this is the first test description", lang="en")
        fd.Description = ODM.Description()
        fd.Description.TranslatedText = [tt]
        fd.ItemGroupRef.append(ODM.ItemGroupRef(ItemGroupOID="ODM.IG.COMMON", Mandatory="Yes", OrderNumber=1))
        fd.ItemGroupRef.append(ODM.ItemGroupRef(ItemGroupOID="ODM.IG.VS_GENERAL", Mandatory="Yes", OrderNumber=2))
        fd.ItemGroupRef.append(ODM.ItemGroupRef(ItemGroupOID="ODM.IG.VS", Mandatory="Yes", OrderNumber=3))
        fd.Alias.append(ODM.Alias(Context="SDTMIG", Name="VS"))
        fd_json = fd.to_json()
        fd_dict = json.loads(fd_json)
        print(fd_dict)
        self.assertEqual(fd_dict["OID"], "ODM.F.VS")
        self.assertDictEqual(fd_dict, self.expected_dict())

    def test_to_dict(self):
        attrs = self.set_formdef_attributes()
        fd = ODM.FormDef(**attrs)
        fd.Description = ODM.Description()
        fd.Description.TranslatedText.append(ODM.TranslatedText(_content="this is the first test description", lang="en"))
        fd.ItemGroupRef.append(ODM.ItemGroupRef(ItemGroupOID="ODM.IG.COMMON", Mandatory="Yes", OrderNumber=1))
        fd.ItemGroupRef.append(ODM.ItemGroupRef(ItemGroupOID="ODM.IG.VS_GENERAL", Mandatory="Yes", OrderNumber=2))
        fd.ItemGroupRef.append(ODM.ItemGroupRef(ItemGroupOID="ODM.IG.VS", Mandatory="Yes", OrderNumber=3))
        fd.Alias.append(ODM.Alias(Context="SDTMIG", Name="VS"))
        fd_dict = fd.to_dict()
        self.assertEqual(fd_dict["OID"], "ODM.F.VS")
        self.assertDictEqual(fd_dict, self.expected_dict())

    def test_to_xml(self):
        attrs = self.set_formdef_attributes()
        fd = ODM.FormDef(**attrs)
        fd.Description = ODM.Description()
        fd.Description.TranslatedText.append(ODM.TranslatedText(_content="this is the first test description", lang="en"))
        fd.ItemGroupRef.append(ODM.ItemGroupRef(ItemGroupOID="ODM.IG.COMMON", Mandatory="Yes", OrderNumber=1))
        fd.ItemGroupRef.append(ODM.ItemGroupRef(ItemGroupOID="ODM.IG.VS_GENERAL", Mandatory="Yes", OrderNumber=2))
        fd.ItemGroupRef.append(ODM.ItemGroupRef(ItemGroupOID="ODM.IG.VS", Mandatory="Yes", OrderNumber=3))
        fd.Alias.append(ODM.Alias(Context="SDTMIG", Name="VS"))
        fd_xml = fd.to_xml()
        self.assertEqual(fd_xml.attrib["OID"], "ODM.F.VS")
        igr = fd_xml.findall("ItemGroupRef")
        self.assertEqual(len(igr), 3)
        self.assertEqual(igr[0].attrib, {"ItemGroupOID": "ODM.IG.COMMON", "Mandatory": "Yes", "OrderNumber": "1"})

    @staticmethod
    def set_formdef_attributes():
        """
        set some FormDef element attributes using test data
        :return: dictionary with FormDef attribute information
        """
        return {"OID": "ODM.F.VS", "Name": "Vital Signs Form", "Repeating": "Yes"}

    @staticmethod
    def expected_dict():
        return {'OID': 'ODM.F.VS', 'Name': 'Vital Signs Form', 'Repeating': 'Yes',
                'ItemGroupRef': [{'ItemGroupOID': 'ODM.IG.COMMON', 'Mandatory': 'Yes', 'OrderNumber': 1},
                                 {'ItemGroupOID': 'ODM.IG.VS_GENERAL', 'Mandatory': 'Yes', 'OrderNumber': 2},
                                 {'ItemGroupOID': 'ODM.IG.VS', 'Mandatory': 'Yes', 'OrderNumber': 3}],
                'Description': {'TranslatedText':
                                [{'_content': 'this is the first test description', 'lang': 'en'}]},
                'Alias': [{'Context': 'SDTMIG', 'Name': 'VS'}]}



from unittest import TestCase
import json
import odmlib.odm_1_3_2.model as ODM


class TestCodeListItem(TestCase):
    def setUp(self) -> None:
        attrs = self._set_attributes()
        self.cli = ODM.CodeListItem(**attrs)

    def test_add_decode(self):
        tt1 = ODM.TranslatedText(_content="Hemoglobin", lang="en")
        decode = ODM.Decode()
        decode.TranslatedText.append(tt1)
        self.cli.Decode = decode
        self.assertEqual(len(self.cli.Decode.TranslatedText), 1)
        self.assertEqual(self.cli.Decode.TranslatedText[0]._content, 'Hemoglobin')

    def test_add_alias(self):
        self.cli.Alias.append(ODM.Alias(Context="nci:ExtCodeID", Name="C64848"))
        self.assertEqual(self.cli.Alias[0].Name, "C64848")

    def test_missing_attribute(self):
        # missing coded value
        attrs1 = {"OrderNumber": 1}
        with self.assertRaises(ValueError):
            cli = ODM.CodeListItem(**attrs1)
        attrs2 = {"CodedValue": "HGB", "OrderNumber": 1}
        cli = ODM.CodeListItem(**attrs2)
        self.assertEqual(cli.CodedValue, "HGB")
        attrs3 = {"DeCodedValue": "HGB", "OrderNumber": 1}
        # invalid attribute
        with self.assertRaises(TypeError):
            cli = ODM.CodeListItem(**attrs3)

    def test_missing_element(self):
        attrs = {"CodedValue": "HGB", "OrderNumber": 1}
        cli = ODM.CodeListItem(**attrs)
        decode = ODM.Decode()
        # this works because TT is set to an empty list - this model violation will be caught with conformance test
        decode.TranslatedText.append(None)
        cli.Decode = decode
        self.assertEqual(len(cli.Decode.TranslatedText), 1)
        # test translating missing element into dict
        with self.assertRaises(AttributeError):
            cli_dict = cli.to_dict()
        with self.assertRaises(AttributeError):
            cli_xml = cli.to_xml()

    def test_to_json(self):
        attrs = self._set_attributes()
        cli = ODM.CodeListItem(**attrs)
        tt1 = ODM.TranslatedText(_content="Hemoglobin", lang="en")
        decode = ODM.Decode()
        decode.TranslatedText.append(tt1)
        cli.Decode = decode
        cli.Alias.append(ODM.Alias(Context="nci:ExtCodeID", Name="C64848"))
        cli_json = cli.to_json()
        cli_dict = json.loads(cli_json)
        print(cli_dict)
        self.assertEqual(cli_dict["CodedValue"], "HGB")
        self.assertDictEqual(cli_dict, self.expected_dict())

    def test_to_dict(self):
        attrs = self._set_attributes()
        cli = ODM.CodeListItem(**attrs)
        tt1 = ODM.TranslatedText(_content="Hemoglobin", lang="en")
        decode = ODM.Decode()
        decode.TranslatedText.append(tt1)
        cli.Decode = decode
        cli.Alias.append(ODM.Alias(Context="nci:ExtCodeID", Name="C64848"))
        cli_dict = cli.to_dict()
        print(cli_dict)
        self.assertDictEqual(cli_dict, self.expected_dict())

    def test_to_xml(self):
        attrs = self._set_attributes()
        cli = ODM.CodeListItem(**attrs)
        tt1 = ODM.TranslatedText(_content="Hemoglobin", lang="en")
        decode = ODM.Decode()
        decode.TranslatedText.append(tt1)
        cli.Decode = decode
        cli.Alias.append(ODM.Alias(Context="nci:ExtCodeID", Name="C64848"))
        cli_xml = cli.to_xml()
        tt = cli_xml.find("Decode/TranslatedText")
        self.assertEqual(tt.text, "Hemoglobin")

    @staticmethod
    def _set_attributes():
        """
        set some CodeListItem element attributes using test data
        :return: (dict) dictionary with CodeListItem attribute information
        """
        return {"CodedValue": "HGB", "OrderNumber": 1}

    @staticmethod
    def expected_dict():
        return {'CodedValue': 'HGB', 'OrderNumber': 1,
                'Decode': {'TranslatedText': [{'lang': 'en', '_content': 'Hemoglobin'}]},
                'Alias': [{'Context': 'nci:ExtCodeID', 'Name': 'C64848'}]}



from unittest import TestCase
import odmlib.ns_registry as NS
import odmlib.odm_1_3_2.model as ODM
import xml.etree.ElementTree as ET
import odmlib.define_loader as DL
import odmlib.loader as LD
import os


class TestNamespaceRegistry(TestCase):
    def setUp(self) -> None:
        NS.NamespaceRegistry(prefix="odm", uri="http://www.cdisc.org/ns/odm/v1.3", is_default=True, is_reset=True)
        NS.NamespaceRegistry(prefix="xs", uri="http://www.w3.org/2001/XMLSchema-instance")
        NS.NamespaceRegistry(prefix="xml", uri="http://www.w3.org/XML/1998/namespace")
        NS.NamespaceRegistry(prefix="xlink", uri="http://www.w3.org/1999/xlink")

    def test_get_ns_prefix(self):
        nsr = NS.NamespaceRegistry(prefix="def", uri="http://www.cdisc.org/ns/def/v2.0")
        def_prefix = nsr.get_prefix_ns_from_uri("http://www.cdisc.org/ns/def/v2.0")
        self.assertEqual("def", def_prefix)

    def test_get_ns_prefix_missing(self):
        nsr = NS.NamespaceRegistry(prefix="def", uri="http://www.cdisc.org/ns/def/v2.0")
        with self.assertRaises(ValueError):
            def_prefix = nsr.get_prefix_ns_from_uri("http://www.cdisc.org/ns/def/v3.0")

    def test_get_ns_entry_dict(self):
        nsr = NS.NamespaceRegistry(prefix="def", uri="http://www.cdisc.org/ns/def/v2.0")
        def_ns_dict = nsr.get_ns_entry_dict("def")
        expected_dict = {"def": "http://www.cdisc.org/ns/def/v2.0"}
        self.assertDictEqual(expected_dict, def_ns_dict)

    def test_get_ns_entry_dict_missing(self):
        nsr = NS.NamespaceRegistry(prefix="def", uri="http://www.cdisc.org/ns/def/v2.0")
        missing_ns_dict = nsr.get_ns_entry_dict("ct")
        expected_dict = {}
        self.assertDictEqual(expected_dict, missing_ns_dict)

    def test_update_registry(self):
        nsr = NS.NamespaceRegistry(prefix="def", uri="http://www.cdisc.org/ns/def/v2.0")
        expected_ns = self.get_ns_defaults()
        expected_ns.update({'odm': 'http://www.cdisc.org/ns/odm/v1.3', 'def': 'http://www.cdisc.org/ns/def/v2.0'})
        self.assertDictEqual(expected_ns, nsr.namespaces)
        nsr = NS.NamespaceRegistry(prefix="nciodm", uri="http://ncicb.nci.nih.gov/xml/odm/EVS/CDISC")
        expected_ns = self.get_ns_defaults()
        expected_ns.update({'odm': 'http://www.cdisc.org/ns/odm/v1.3', 'def': 'http://www.cdisc.org/ns/def/v2.0',
                            'nciodm': 'http://ncicb.nci.nih.gov/xml/odm/EVS/CDISC'})
        self.assertDictEqual(expected_ns, nsr.namespaces)

    def test_use_existing_registry(self):
        expected_ns = self.get_ns_defaults()
        expected_ns.update({'odm': 'http://www.cdisc.org/ns/odm/v1.3'})
        nsr = NS.NamespaceRegistry()
        self.assertDictEqual(expected_ns, nsr.namespaces)

    def test_remove_registry_entry(self):
        # tests a singleton so each test builds on the previous one
        NS.NamespaceRegistry(prefix="def", uri="http://www.cdisc.org/ns/def/v2.0")
        nsr = NS.NamespaceRegistry(prefix="nciodm", uri="http://ncicb.nci.nih.gov/xml/odm/EVS/CDISC")
        expected_ns = self.get_ns_defaults()
        expected_ns.update({'odm': 'http://www.cdisc.org/ns/odm/v1.3',
                            'def': 'http://www.cdisc.org/ns/def/v2.0',
                            'nciodm': 'http://ncicb.nci.nih.gov/xml/odm/EVS/CDISC'})
        self.assertDictEqual(expected_ns, nsr.namespaces)
        nsr.remove_registry_entry("def")
        expected_ns = self.get_ns_defaults()
        expected_ns.update({'odm': 'http://www.cdisc.org/ns/odm/v1.3',
                            'nciodm': 'http://ncicb.nci.nih.gov/xml/odm/EVS/CDISC'})
        self.assertDictEqual(expected_ns, nsr.namespaces)

    def test_default_ns(self):
        NS.NamespaceRegistry(prefix="def", uri="http://www.cdisc.org/ns/def/v2.0")
        nsr = NS.NamespaceRegistry(prefix="nciodm", uri="http://ncicb.nci.nih.gov/xml/odm/EVS/CDISC")
        expected_ns = self.get_ns_defaults()
        expected_ns.update({'odm': 'http://www.cdisc.org/ns/odm/v1.3', 'def': 'http://www.cdisc.org/ns/def/v2.0',
                            'nciodm': 'http://ncicb.nci.nih.gov/xml/odm/EVS/CDISC'})
        self.assertDictEqual(expected_ns, nsr.namespaces)
        self.assertDictEqual({'odm': 'http://www.cdisc.org/ns/odm/v1.3'}, nsr.default_namespace)

    def test_odm_ns_entries(self):
        NS.NamespaceRegistry(prefix="def", uri="http://www.cdisc.org/ns/def/v2.0")
        nsr = NS.NamespaceRegistry(prefix="nciodm", uri="http://ncicb.nci.nih.gov/xml/odm/EVS/CDISC")
        expected_ns = self.get_ns_defaults()
        expected_ns.update({'odm': 'http://www.cdisc.org/ns/odm/v1.3', 'def': 'http://www.cdisc.org/ns/def/v2.0',
                            'nciodm': 'http://ncicb.nci.nih.gov/xml/odm/EVS/CDISC'})
        self.assertDictEqual(expected_ns, nsr.namespaces)
        odm_ns = nsr.get_odm_namespace_entries()
        expected_list = ['xmlns=http://www.cdisc.org/ns/odm/v1.3',
                         'xmlns:xs=http://www.w3.org/2001/XMLSchema-instance',
                         'xmlns:xml=http://www.w3.org/XML/1998/namespace',
                         'xmlns:xlink=http://www.w3.org/1999/xlink',
                         'xmlns:def=http://www.cdisc.org/ns/def/v2.0',
                         'xmlns:nciodm=http://ncicb.nci.nih.gov/xml/odm/EVS/CDISC']
        self.assertListEqual(expected_list, odm_ns)

    def test_prefix_without_url(self):
        expected_ns = self.get_ns_defaults()
        expected_ns.update({'odm': 'http://www.cdisc.org/ns/odm/v1.3'})
        nsr = NS.NamespaceRegistry(prefix="def")
        self.assertDictEqual(expected_ns, nsr.namespaces)
        nsr = NS.NamespaceRegistry(uri='http://www.cdisc.org/ns/def/v2.0')
        self.assertDictEqual(expected_ns, nsr.namespaces)

    def test_bad_ns_url(self):
        with self.assertRaises(ValueError):
            nsr = NS.NamespaceRegistry(prefix="def", uri="www.cdisc.org/ns/def/v2.0")

    def test_odm_ns_attributes(self):
        odm_test_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'test_odm_namespace.xml')
        NS.NamespaceRegistry(prefix="def", uri="http://www.cdisc.org/ns/def/v2.0")
        NS.NamespaceRegistry(prefix="nciodm", uri="http://ncicb.nci.nih.gov/xml/odm/EVS/CDISC")
        NS.NamespaceRegistry(prefix="xsi", uri="http://www.w3.org/2001/XMLSchema-instance")
        root = self.add_root()
        odm_xml = root.to_xml()
        nsr = NS.NamespaceRegistry()
        nsr.set_odm_namespace_attributes(odm_xml)
        tree = ET.ElementTree(odm_xml)
        tree.write(odm_test_file, xml_declaration=True, encoding='utf-8', method='xml', short_empty_elements=True)
        self.assertEqual("ODM.MDV.TEST.001", root.FileOID)
        loader = LD.ODMLoader(DL.XMLDefineLoader())
        odm_root = loader.open_odm_document(odm_test_file)
        self.assertEqual("ODM.MDV.TEST.001", odm_root.attrib["FileOID"])
        odm = loader.create_odmlib(odm_root)
        self.assertEqual("ODM.MDV.TEST.001", odm.FileOID)
        nsr2 = NS.NamespaceRegistry()
        namespaces = nsr2.get_odm_namespace_entries()
        expected_ns_list = ['xmlns=http://www.cdisc.org/ns/odm/v1.3',
                            'xmlns:xs=http://www.w3.org/2001/XMLSchema-instance',
                            'xmlns:xml=http://www.w3.org/XML/1998/namespace',
                            'xmlns:xlink=http://www.w3.org/1999/xlink',
                            'xmlns:def=http://www.cdisc.org/ns/def/v2.0',
                            'xmlns:nciodm=http://ncicb.nci.nih.gov/xml/odm/EVS/CDISC',
                            'xmlns:xsi=http://www.w3.org/2001/XMLSchema-instance']
        self.assertListEqual(namespaces, expected_ns_list)




    def add_root(self):
        attrs = self.get_root_attributes()
        root = ODM.ODM(**attrs)
        return root

    def get_root_attributes(self):
        attrs = {"FileOID": "ODM.MDV.TEST.001", "AsOfDateTime": "2020-07-13T00:13:51.309617+00:00",
                 "CreationDateTime": "2020-07-13T00:13:51.309617+00:00", "ODMVersion": "1.3.2", "FileType": "Snapshot",
                 "Originator": "RDS", "SourceSystem": "ODMLib", "SourceSystemVersion": "0.1"}
        return attrs

    def get_ns_defaults(self):
        return {
            'xs': 'http://www.w3.org/2001/XMLSchema-instance',
            'xml': 'http://www.w3.org/XML/1998/namespace',
            'xlink': 'http://www.w3.org/1999/xlink'
        }



from unittest import TestCase
import odmlib.odm_loader as OL
import odmlib.loader as LD
import os
import json


class TestJsonFromXml(TestCase):
    def setUp(self) -> None:
        self.odm_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'cdash-odm-test.xml')
        self.loader = LD.ODMLoader(OL.XMLODMLoader())

    def test_open_odm_document(self):
        root = self.loader.open_odm_document(self.odm_file)
        elem_name = root.tag[root.tag.find('}') + 1:]
        self.assertEqual("ODM", elem_name)
        self.assertEqual("CDASH_File_2011-10-24", root.attrib["FileOID"])

    def test_json_from_xml(self):
        root = self.loader.open_odm_document(self.odm_file)
        odm = self.loader.create_odmlib(root)
        odm_json = odm.to_json()
        odm_json_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'cdash_odm_test.json')
        with open(odm_json_file, "w") as odm_in:
            odm_in.write(odm_json)
        odm_dict = json.loads(odm_json)
        self.assertEqual(odm_dict["Study"][0]["MetaDataVersion"][0]["ItemDef"][0]["OID"], "ODM.IT.Common.StudyID")



import unittest
import os
import odmlib.odm_parser as P


class TestOdmParserMetaData(unittest.TestCase):
    def setUp(self) -> None:
        self.odm_file_1 = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data', 'cdash_odm_test.json')
        with open(self.odm_file_1, "r", encoding="utf-8") as f:
            self.odm_string = f.read()
        self.parser = P.ODMJSONStringParser(self.odm_string)
        self.root = self.parser.parse()
        self.mdv = self.parser.MetaDataVersion()

    def test_MetaDataVersion(self):
        self.assertTrue(isinstance(self.mdv, list))
        self.assertEqual(self.mdv[0]["Name"], "TRACE-XML MDV")
        self.assertEqual(self.mdv[0]["OID"], "MDV.TRACE-XML-ODM-01")
        self.assertEqual(8, len(self.mdv[0]))

    def test_Protocol(self):
        protocol = self.mdv[0]["Protocol"]
        self.assertIsInstance(protocol['StudyEventRef'], list)
        self.assertEqual(protocol['StudyEventRef'][0]["StudyEventOID"], "BASELINE")

    def test_StudyEventDef(self):
        sed = self.mdv[0]["StudyEventDef"]
        self.assertEqual(sed[0]["OID"], "BASELINE")
        self.assertEqual(sed[0]["Name"], "Baseline Visit")

    def test_FormRef(self):
        fr = self.mdv[0]["StudyEventDef"][0]["FormRef"]
        self.assertEqual(fr[0]["FormOID"], "ODM.F.DM")
        fr_list = ["ODM.F.DM", "ODM.F.VS", "ODM.F.AE"]
        self.assertListEqual(fr_list, [f["FormOID"] for f in fr])

    def test_FormDef(self):
        fd = self.mdv[0]["FormDef"]
        self.assertEqual(fd[0]["OID"], "ODM.F.DM")
        self.assertEqual(fd[2]["Name"], "Adverse Event")

    def test_ItemGroupDef(self):
        ig = self.mdv[0]["ItemGroupDef"]
        self.assertEqual(len(ig), 7)
        self.assertEqual('Demographics', ig[1]['Name'])
        self.assertEqual('ODM.IG.DM', ig[1]['OID'])
        self.assertEqual(ig[3]["OID"], "ODM.IG.VS")


if __name__ == '__main__':
    unittest.main()



import odmlib.descriptor as DESC
import odmlib.typed as T
import odmlib.ns_registry as NS
import odmlib.oid_index as IDX
from collections import OrderedDict
import json
import xml.etree.ElementTree as ET


class ODMMeta(type):
    @classmethod
    def __prepare__(cls, name, bases):
        """ preserves the order of declarations in each class """
        return OrderedDict()

    def __new__(cls, clsname, bases, clsdict):
        # variables created in classes become the class attributes
        # fields = [key for key, val in clsdict.items() if isinstance(val, (DESC.Descriptor, ODMMeta))]
        clsdict["_fields"] = []
        clsdict["_elems"] = {}
        clsdict["_attrs"] = {}
        clsdict["_attr_ns"] = {}
        for key, val in clsdict.items():
            if isinstance(val, (T.ODMObject, T.ODMListObject)):
                clsdict["_elems"][key] = val
                clsdict[key].name = key
                clsdict["_fields"].append(key)
            elif isinstance(val, (DESC.Descriptor, ODMMeta)):
                clsdict["_attrs"][key] = val
                clsdict[key].name = key
                clsdict["_fields"].append(key)
                if val.namespace != "odm":
                    clsdict["_attr_ns"][key] = val.namespace

        # for name in fields:
        #     clsdict[name].name = name

        #clsdict["_fields"] = fields
        # the default class namespace is odm
        if "namespace" not in clsdict:
            clsdict["namespace"] = "odm"
        # elems = {key: val for key, val in clsdict.items() if isinstance(val, (T.ODMObject, T.ODMListObject))}
        # clsdict["_elems"] = elems
        # add attribute non-default namespaces
        # ns = {key: val.namespace for key, val in clsdict.items() if isinstance(val, (DESC.Descriptor, ODMMeta))
        #       if val.namespace != "odm"}
        # clsdict["_attr_ns"] = ns

        clsobj = super().__new__(cls, clsname, bases, dict(clsdict))
        return clsobj


class ODMWriter:

    @staticmethod
    def write_odm(odm_file, odm_elem):
        """
        after converting ODMLIB to ElementTree, write the ElementTree to an ODM file
        :param odm_file: path and file to write the ODM XML
        :param odm_elem: Element object to write to ODM (presumably an ODM root)
        """
        tree = ET.ElementTree(odm_elem)
        root = tree.getroot()
        # workaround for elementtree NS bug - NamespaceRegistry assumes at least 1 default NS has been set
        nsr = NS.NamespaceRegistry()
        nsr.set_odm_namespace_attributes(root)
        tree.write(odm_file, xml_declaration=True, encoding='utf-8', method='xml', short_empty_elements=True)


class ODMElement(metaclass=ODMMeta):
    def __init__(self, **kwargs):
        """
        abstract ODM element class used to set the properties of any ODM object
        """
        for name, val in kwargs.items():
            if name not in self.__class__.__dict__.keys():
                # strip out non-default elementtree namespaces from the XML to work with just the name e.g. xml:lang
                if "}" in name:
                    name = name[name.find('}') + 1:]
                else:
                    raise TypeError(f"Unknown keyword argument {name} in {self.__class__.__name__}")
            setattr(self, name, val)
        for attr, obj in self.__class__.__dict__.items():
            if isinstance(obj, DESC.Descriptor) and (not isinstance(obj, T.ODMObject)) and (attr not in self.__dict__) and obj.required:
                raise ValueError(f"Missing required keyword argument {attr} in {self.__class__.__name__}")

    def __setattr__(self, key, value):
        """ ensure the object being added is a type that belongs to the class """
        if not hasattr(self, key):
            raise TypeError(f"Assignment error: {self.__class__.__name__} does not have a defined attribute {key}")
        super().__setattr__(key, value)

    def to_json(self):
        """
        transforms odmlib hierarchy into a dict and converts that to JSON and returns it

        :return: JSON representation of odmlib hierarchy
        """
        return json.dumps(self.to_dict())

    def to_xml(self, parent_elem=None, top_elem=None):
        """
        generates ElementTree XML from the odmlib object hierarchy

        :param parent_elem: (obj) Element that is the parent of the element to be created
        :param top_elem: (obj) Topmost element that has been created
        :return: top_elem: (obj) Topmost element created during the recursive calls
        """
        # create attributes
        attrs = {}
        for attr, obj in self.__dict__.items():
            if not isinstance(obj, (ODMElement, list)) and attr != "_content" and obj is not None:
                # add namespace if not the default namespace
                if attr in self.__class__.__dict__["_attr_ns"]:
                    attrs[self.__class__.__dict__["_attr_ns"][attr] + ":" + attr] = str(obj)
                else:
                    attrs[attr] = str(obj)

        # create element
        if isinstance(parent_elem, ET.Element):
            odm_elem = ET.SubElement(parent_elem, self.__class__.__name__ if self.namespace == "odm" else self.namespace + ":" + self.__class__.__name__, attrs)
        else:
            odm_elem = ET.Element(self.__class__.__name__ if self.namespace == "odm" else self.namespace + ":" + self.__class__.__name__, attrs)
            top_elem = odm_elem
        # add text to element if it exists
        if "_content" in self.__dict__:
            odm_elem.text = self.__dict__["_content"]
        for name, obj in self.__dict__.items():
            # process each element in a list of ELEMENTS
            if isinstance(obj, list) and obj:
                for o in obj:
                    o.to_xml(odm_elem, top_elem)
            elif isinstance(obj, ODMElement):
                obj.to_xml(odm_elem, top_elem)
        return top_elem

    def to_dict(self):
        """
        transform odmlib object hierarchy into a Python dictionary and return it

        :return: dictionary serialization of odmlib hierarchy
        """
        # Note: namespaces used in the XML serialization are not part of the dictionary or json serializations
        property_dict = {}
        odm_content = {attr: obj for attr, obj in self.__dict__.items() if attr not in ["_fields", "_attr_ns", "_elems", "_attrs"]}
        for attr, obj in odm_content.items():
            if isinstance(obj, ODMElement):
                property_dict[attr] = obj.to_dict()                    # element
            elif isinstance(obj, list):
                property_dict[attr] = [o.to_dict() for o in obj]       # list of ELEMENTS
            elif obj is not None:
                property_dict[attr] = obj                              # attributes
        return property_dict

    def __repr__(self):
        args = ", ".join(name for name in self._fields)
        return type(self).__name__ + "(" + args + ")"

    def __str__(self):
        if "_content" in self._fields and self._content:
            return self._content
        elif "OID" in self._fields and self.OID:
            return type(self).__name__ + "(OID=" + self.OID + ")"
        else:
            args = ", ".join(name for name in self._fields)
            return type(self).__name__ + "(" + args + ")"

    def find(self, obj_name, attr, val):
        """
        return an odmlib object for object type obj_name where the attribute attr value equals val

        :param obj_name: text name of the ODM Element (case sensitive)
        :param attr: text name of the ODM Element Attribute (case sensitive)
        :param val: attribute value to search for
        :return: an odmlib object of the ODM element with the first time the attribute value matches val
        """
        obj_list = eval("self." + obj_name)
        for o in obj_list:
            if o.__dict__[attr] == val:
                return o
        return None

    def write_xml(self, odm_file, odm_writer=ODMWriter):
        """
        write the odmlib hierarchy as an XML file

        :param odm_file: string ODM filename and path
        :param odm_writer: object used to write the elementree XML to a file
        """
        odm_elem = self.to_xml()
        odm_writer = odm_writer()
        odm_writer.write_odm(odm_file, odm_elem)

    def write_json(self, odm_file):
        """
        write the odmlib hierarchy as a JSON file

        :param odm_file: string ODM filename and path
        """
        with open(odm_file, 'w') as outfile:
            json.dump(self.to_dict(), outfile)

    def build_oid_index(self):
        idx = IDX.OIDIndex()
        self._init_oid_index(idx)
        return idx

    def _init_oid_index(self, idx):
        """
        for odmlib object, loads all OIDs into a dict that functions as an OID index

        :return oid_index: object that provices a dictionary lookup based on OID
        """
        odm_content = {attr: obj for attr, obj in self.__dict__.items() if attr not in ["_fields", "_attr_ns", "_elems", "_attrs"]}
        for attr, obj in odm_content.items():
            if isinstance(obj, ODMElement):
                obj._init_oid_index(idx)                    # element
            elif isinstance(obj, list):
                for o in obj:
                    o._init_oid_index(idx)                  # list of ELEMENTS
            else:
                if attr == "OID" or "OID" in attr:
                    idx.add_oid(obj, self)
        return

    def verify_oids(self, oid_checker):
        """
        checks all the OIDs for uniqueness and Def/Ref integrity; oid_checker throws a ValueError on failure

        :param oid_checker: object that performs that checks OID uniqueness and Def/Ref checks
        """
        self._init_oid_check(oid_checker)
        oid_checker.check_oid_refs()

    def _init_oid_check(self, oid_checker):
        """
        for odmlib object, loads all OIDs and checks them for uniqueness; throws an error if uniqueness check fails

        :param oid_checker: object used to check OIDs for uniqueness and Def/Ref check
        """
        odm_content = {attr: obj for attr, obj in self.__dict__.items() if attr not in ["_fields", "_attr_ns", "_elems", "_attrs"]}
        for attr, obj in odm_content.items():
            if isinstance(obj, ODMElement):
                obj._init_oid_check(oid_checker)                    # element
            elif isinstance(obj, list):
                for o in obj:
                    o._init_oid_check(oid_checker)                  # list of ELEMENTS
            else:
                if attr == "OID":
                    oid_checker.add_oid(obj, self.__class__.__name__)
                elif "OID" in attr:
                    oid_checker.add_oid_ref(obj, attr)
        return

    def verify_conformance(self, validator):
        """
        uses validator object to check object for conformance with the model

        :param validator: object that validates the odmlib object against the model
        """
        doc_dict = self.to_dict()
        result = validator.verify_conformance(doc_dict, type(self).__name__)
        return result

    def verify_order(self):
        odm_content = {attr: obj for attr, obj in self.__dict__.items() if attr not in ["_fields", "_attr_ns", "_elems", "_attrs"]}
        obj_list = [key for key in list(self.__dict__.keys()) if key != "_content" and key not in self._attrs]
        elem_list = [elem for elem in self._elems if elem in obj_list]
        if obj_list != elem_list:
            raise ValueError(f"The order of elements in {self.__class__.__name__} should be {', '.join([key for key in self._elems.keys()])}")
        for attr, obj in odm_content.items():
            if isinstance(obj, ODMElement):
                obj.verify_order()
            elif isinstance(obj, list):
                for o in obj:
                    o.verify_order()
        return True

    def reorder_object(self):
        ordered_obj = OrderedDict()
        for model_elem_name, model_elem_obj in self._elems.items():
            if model_elem_name in self.__dict__:
                obj = self.__dict__.pop(model_elem_name)
                ordered_obj[model_elem_name] = obj
        for name, elem in ordered_obj.items():
            self.__dict__[name] = elem




import unittest
import odmlib.odm_1_3_2.model as ODM
import odmlib.odm_1_3_2.rules.metadata_schema as METADATA
import datetime
import odmlib.typed as T
import odmlib.odm_element as OE


class TestText(OE.ODMElement):
    Name = T.String(required=True)
    Label = T.Sized(required=True, max_length=4)
    Range = T.Regex(required=False, pat="[0-4]-[0-9]")
    Link = T.Url(required=False)
    TestDateTime = T.IncompleteDateTimeString(required=False)
    TestDate = T.IncompleteDateString(required=False)
    TestTime = T.IncompleteTimeString(required=False)


class TestPartialDate(OE.ODMElement):
    Name = T.String(required=True)
    TestDateTime = T.PartialDateTimeString(required=False)
    TestDate = T.PartialDateString(required=False)
    TestTime = T.PartialTimeString(required=False)


class TestTyped(unittest.TestCase):
    def test_integer_type(self):
        # order number is an integer - test as an integer
        attrs = {"CodedValue": "HGB", "OrderNumber": 1}
        cli = ODM.CodeListItem(**attrs)
        self.assertEqual(cli.OrderNumber, 1)
        # test integer as string
        attrs = {"CodedValue": "HGB", "OrderNumber": "1"}
        cli = ODM.CodeListItem(**attrs)
        self.assertEqual(cli.OrderNumber, 1)
        # test integer as float
        attrs = {"CodedValue": "HGB", "OrderNumber": 1.1}
        with self.assertRaises(TypeError):
            cli = ODM.CodeListItem(**attrs)
        # test non-integer as
        attrs = {"CodedValue": "HGB", "OrderNumber": "a"}
        with self.assertRaises(TypeError):
            cli = ODM.CodeListItem(**attrs)

    def test_float_type(self):
        # rank is a float - test as a float
        attrs = {"CodedValue": "HGB", "Rank": 1.1}
        cli = ODM.CodeListItem(**attrs)
        self.assertEqual(cli.Rank, 1.1)
        # test float as string
        attrs = {"CodedValue": "HGB", "Rank": "1.1"}
        cli = ODM.CodeListItem(**attrs)
        self.assertEqual(cli.Rank, 1.1)
        # test integer as float
        attrs = {"CodedValue": "HGB", "Rank": 1}
        cli = ODM.CodeListItem(**attrs)
        self.assertEqual(cli.Rank, 1.0)
        # test non-float as string
        attrs = {"CodedValue": "HGB", "Rank": "a"}
        with self.assertRaises(TypeError):
            cli = ODM.CodeListItem(**attrs)
        # test object as float
        clr = ODM.CodeListRef(CodeListOID="CL.TEST")
        attrs = {"CodedValue": "HGB", "Rank": clr}
        with self.assertRaises(TypeError):
            cli = ODM.CodeListItem(**attrs)

    def test_positive_type(self):
        # testing Length as a positive integer
        attrs = {"OID": "ODM.IT.AE.AEYN", "Name": "Any AEs?", "DataType": "text", "Length": 1, "SASFieldName": "AEYN",
                "SDSVarName": "AEYN", "Origin": "CRF", "Comment": "Data management field"}
        item = ODM.ItemDef(**attrs)
        self.assertEqual(item.Length, 1)
        # testing Length = 0 as a positive integer
        attrs["Length"] = 0
        with self.assertRaises(TypeError):
            item = ODM.ItemDef(**attrs)
        # testing Length = -1 as a positive integer
        attrs["Length"] = -1
        with self.assertRaises(TypeError):
            item = ODM.ItemDef(**attrs)
        # testing Length = float as a positive integer
        attrs["Length"] = 1.0
        with self.assertRaises(TypeError):
            item = ODM.ItemDef(**attrs)

    def test_non_negative_type(self):
        # testing SignificantDigits as a non-negative integer
        attrs = {"OID": "ODM.IT.AE.AEYN", "Name": "Any AEs?", "DataType": "text", "Length": 1, "SASFieldName": "AEYN",
                "SDSVarName": "AEYN", "Origin": "CRF", "Comment": "Data management field", "SignificantDigits": 1}
        item = ODM.ItemDef(**attrs)
        self.assertEqual(item.SignificantDigits, 1)
        # testing SignificantDigits = 0 as a non-negative integer
        attrs["SignificantDigits"] = 0
        item = ODM.ItemDef(**attrs)
        self.assertEqual(item.SignificantDigits, 0)
        # testing SignificantDigits = -1 as a non-negative integer
        attrs["SignificantDigits"] = -1
        with self.assertRaises(TypeError):
            item = ODM.ItemDef(**attrs)
        # testing SignificantDigits = float as a non-negative integer
        attrs["SignificantDigits"] = 1.0
        with self.assertRaises(TypeError):
            item = ODM.ItemDef(**attrs)

    def test_add_undefined_content(self):
        # use conformance rules to check for unknown objects being added after creation
        attrs = {"OID": "ODM.IT.AE.AEYN", "Name": "Any AEs?", "DataType": "text", "Length": 1, "SASFieldName": "AEYN",
                "SDSVarName": "AEYN", "Origin": "CRF", "Comment": "Data management field", "SignificantDigits": 1}
        item = ODM.ItemDef(**attrs)
        clr = ODM.CodeListRef(CodeListOID="CL.TEST")
        with self.assertRaises(TypeError):
            item.NewCodeList = clr
        # cannot add unknown attributes or elements during creation, but can afterwards; catch with conformance checks
        attrs = {"OID": "ODM.IT.AE.AEYN", "Name": "Any AEs?", "DataType": "text", "Length": 1, "SASFieldName": "AEYN",
                "SDSVarName": "AEYN", "Origin": "CRF", "Comment": "Data management field", "SignificantDigits": 1}
        item = ODM.ItemDef(**attrs)
        with self.assertRaises(TypeError):
            item.InSignificantDigits = 1
        # can add all objects during creation and they are validated
        attrs = {"OID": "ODM.IT.AE.AEYN", "Name": "Any AEs?", "DataType": "text", "Length": 1, "SASFieldName": "AEYN",
                "SDSVarName": "AEYN", "Origin": "CRF", "Comment": "Data management field", "CodeListRef": clr}
        item = ODM.ItemDef(**attrs)
        self.assertEqual(item.CodeListRef, clr)
        attrs["CodeListReference"] = clr
        with self.assertRaises(TypeError):
            # test shows that ItemDef fails the schema check because it has an unknown CodeListReference object
            item = ODM.ItemDef(**attrs)
        self.assertEqual(item.CodeListRef, clr)
        # test adding the wrong type of object to a list
        attrs = {"OID": "ODM.IT.AE.AEYN", "Name": "Any AEs?", "DataType": "text", "Length": 1, "SASFieldName": "AEYN",
                "SDSVarName": "AEYN", "Origin": "CRF", "Comment": "Data management field", "Alias": [clr]}
        with self.assertRaises(TypeError):
            item = ODM.ItemDef(**attrs)
        # test adding the wrong type of object to a list
        alias = ODM.Alias(Context="CDASH", Name="AEYN")
        attrs = {"OID": "ODM.IT.AE.AEYN", "Name": "Any AEs?", "DataType": "text", "Length": 1, "SASFieldName": "AEYN",
                "SDSVarName": "AEYN", "Origin": "CRF", "Comment": "Data management field", "Alias": alias}
        with self.assertRaises(TypeError):
            item = ODM.ItemDef(**attrs)
        # assign wrong type of object to an element
        attrs = {"OID": "ODM.IT.AE.AEYN", "Name": "Any AEs?", "DataType": "text", "Length": 1, "SASFieldName": "AEYN",
                "SDSVarName": "AEYN", "Origin": "CRF", "Comment": "Data management field", "CodeListRef": alias}
        with self.assertRaises(TypeError):
            item = ODM.ItemDef(**attrs)

    def test_email_type(self):
        # good email
        email = ODM.Email(_content="swhume@gmail.com")
        self.assertEqual(email._content, "swhume@gmail.com")
        # missing @
        with self.assertRaises(ValueError):
            email = ODM.Email(_content="swhumegmail.com")
        # includes a space
        with self.assertRaises(ValueError):
            email = ODM.Email(_content="swhume @gmail.com")

    def test_sasformat_type(self):
        attrs = {"OID": "ODM.CL.LBTESTCD", "Name": "Laboratory Test Code", "DataType": "text",
                 "SASFormatName": "$AETERM"}
        cl = ODM.CodeList(**attrs)
        self.assertEqual(cl.SASFormatName, "$AETERM")
        # start with an underscore
        attrs["SASFormatName"] = "_AETERM"
        cl = ODM.CodeList(**attrs)
        self.assertEqual(cl.SASFormatName, "_AETERM")
        # start with a number
        attrs["SASFormatName"] = "9AETERM"
        with self.assertRaises(ValueError):
            cl = ODM.CodeList(**attrs)
        # end with a symbol
        attrs["SASFormatName"] = "AETERM%"
        with self.assertRaises(ValueError):
            cl = ODM.CodeList(**attrs)
        # test with none - does not throw an error when value is None
        attrs["SASFormatName"] = None
        cl = ODM.CodeList(**attrs)
        self.assertIsNone(cl.SASFormatName)

    def test_sasname_type(self):
        # valid test
        attrs = {"OID": "ODM.IT.AE.AEYN", "Name": "Any AEs?", "DataType": "text", "Length": 1, "SASFieldName": "AEYN",
                "SDSVarName": "AEYN", "Origin": "CRF", "Comment": "Data management field", "SignificantDigits": 1}
        item = ODM.ItemDef(**attrs)
        self.assertEqual(item.SASFieldName, "AEYN")
        # start with an underscore
        attrs["SASFieldName"] = "_AETERM"
        item = ODM.ItemDef(**attrs)
        self.assertEqual(item.SASFieldName, "_AETERM")
        # start with a number
        attrs["SASFieldName"] = "9AETERM"
        with self.assertRaises(ValueError):
            item = ODM.ItemDef(**attrs)
        # end with a symbol
        attrs["SASFieldName"] = "AETERM%"
        with self.assertRaises(ValueError):
            item = ODM.ItemDef(**attrs)
        # test with none - does not throw an error when value is None
        attrs["SASFieldName"] = None
        item = ODM.ItemDef(**attrs)
        self.assertIsNone(item.SASFieldName)

    def test_datetime_type(self):
        attrs = {"FileOID": "ODM.MDV.TEST.001", "Granularity": "Metadata",
                 "AsOfDateTime": "2020-07-13T00:13:51.309617+00:00",
                 "CreationDateTime": "2020-07-13T00:13:51.309617+00:00",
                 "ODMVersion": "1.3.2", "FileType": "Snapshot", "Originator": "RDS", "SourceSystem": "ODMLib"}
        odm = ODM.ODM(**attrs)
        self.assertEqual(odm.AsOfDateTime, "2020-07-13T00:13:51.309617+00:00")
        # derived current datetime in ISO formate from datetime library
        attrs["AsOfDateTime"] = self.set_datetime()
        odm = ODM.ODM(**attrs)
        self.assertIsInstance(odm, ODM.ODM)
        # date only
        attrs["AsOfDateTime"] = "2020-07-13"
        with self.assertRaises(ValueError):
            odm = ODM.ODM(**attrs)
        # basic time
        attrs["AsOfDateTime"] = "2020-07-13T00:13:51"
        odm = ODM.ODM(**attrs)
        self.assertEqual(odm.AsOfDateTime, "2020-07-13T00:13:51")
        # out of range time
        attrs["AsOfDateTime"] = "2020-07-13T00:13:61"
        with self.assertRaises(ValueError):
            odm = ODM.ODM(**attrs)

    def set_datetime(self):
        """return the current datetime in ISO 8601 format"""
        return datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc).isoformat()

    def test_string_type(self):
        attrs = {"OID": "ODM.IT.AE.AEYN", "Name": "Any AEs?", "DataType": "text", "Length": 1, "SASFieldName": "AEYN",
                "SDSVarName": "AEYN", "Origin": "CRF", "Comment": "Data management field", "SignificantDigits": 1}
        item = ODM.ItemDef(**attrs)
        self.assertEqual(item.Origin, "CRF")
        # integer
        attrs["Origin"] = 1
        with self.assertRaises(TypeError):
            item = ODM.ItemDef(**attrs)
        attrs["Origin"] = None
        item = ODM.ItemDef(**attrs)
        self.assertIsNone(item.Origin)

    def test_sized_type(self):
        sized = TestText(Name="AETERM", Label="help")
        self.assertEqual(sized.Label, "help")
        # too long
        with self.assertRaises(ValueError):
            sized = TestText(Name="AETERM", Label="Verbatim Term")

    def test_regex_type(self):
        regex = TestText(Name="AETERM", Label="help", Range="4-9")
        self.assertEqual(regex.Range, "4-9")
        # not a match
        with self.assertRaises(ValueError):
            regex = TestText(Name="AETERM", Label="help", Range="6-7")

    def test_url_type(self):
        url = TestText(Name="AETERM", Label="help", Range="4-9", Link="https://cdisc.org")
        self.assertEqual(url.Link, "https://cdisc.org")
        # not a match
        with self.assertRaises(ValueError):
            url = TestText(Name="AETERM", Label="help", Range="4-9", Link="cdisc.org")

    def test_filename_type(self):
        archive = ODM.ArchiveLayout(OID="AL.AECRF", PdfFileName="ae_annotated_crf.pdf")
        self.assertEqual(archive.PdfFileName, "ae_annotated_crf.pdf")
        # with path
        with self.assertRaises(ValueError):
            archive = ODM.ArchiveLayout(OID="AL.AECRF", PdfFileName="c:\\users\\shume\\ae_annotated_crf.pdf")
        # with space
        archive = ODM.ArchiveLayout(OID="AL.AECRF", PdfFileName="ae_annotated crf.pdf")
        self.assertEqual(archive.PdfFileName, "ae_annotated crf.pdf")
        # with invalid character
        with self.assertRaises(ValueError):
            archive = ODM.ArchiveLayout(OID="AL.AECRF", PdfFileName="ae_annotated>crf.pdf")

    def test_incomplete_datetime_type(self):
        # valid incomplete datetime 2004-12-01T01:01:01Z and 2004---15T-:05:-
        idatetime = TestText(Name="AETERM", Label="help", Range="4-9", TestDateTime="2004-12-01T01:01:01Z")
        self.assertEqual(idatetime.TestDateTime, "2004-12-01T01:01:01Z")
        idatetime = TestText(Name="AETERM", Label="help", Range="4-9", TestDateTime="2004---15T-:05:-")
        self.assertEqual(idatetime.TestDateTime, "2004---15T-:05:-")
        # invalid incomplete datetime
        with self.assertRaises(ValueError):
            idatetime = TestText(Name="AETERM", Label="help", Range="4-7", TestDateTime="2004---15T-:05")

    def test_incomplete_date_type(self):
        # valid incomplete date 2004-12-20 and 2004---15 and ------
        idate = TestText(Name="AETERM", Label="help", Range="4-9", TestDate="2004-12-20")
        self.assertEqual(idate.TestDate, "2004-12-20")
        idate = TestText(Name="AETERM", Label="help", Range="4-9", TestDate="2004---15")
        self.assertEqual(idate.TestDate, "2004---15")
        idate = TestText(Name="AETERM", Label="help", Range="4-7", TestDate="-----")
        self.assertEqual(idate.TestDate, "-----")
        # invalid incomplete date
        with self.assertRaises(ValueError):
            idate = TestText(Name="AETERM", Label="help", Range="4-7", TestDate="----32")

    def test_incomplete_time_type(self):
        # valid incomplete time 2004-12-01T01:01:01Z and 2004---15T-:05:-
        itime = TestText(Name="AETERM", Label="help", Range="4-9", TestTime="01:01:01Z")
        self.assertEqual(itime.TestTime, "01:01:01Z")
        itime = TestText(Name="AETERM", Label="help", Range="4-9", TestTime="-:05:-")
        self.assertEqual(itime.TestTime, "-:05:-")
        # invalid incomplete datetime
        with self.assertRaises(ValueError):
            itime = TestText(Name="AETERM", Label="help", Range="4-7", TestTime="-:05")

    def test_partial_datetime_type(self):
        # valid partial datetime
        pdatetime = TestPartialDate(Name="AETERM", TestDateTime="2004-12-01T01:01:01Z")
        self.assertEqual(pdatetime.TestDateTime, "2004-12-01T01:01:01Z")
        pdatetime = TestPartialDate(Name="AETERM", TestDateTime="2004-12")
        self.assertEqual(pdatetime.TestDateTime, "2004-12")
        pdatetime = TestPartialDate(Name="AETERM", TestDateTime="2004-12-05T12")
        self.assertEqual(pdatetime.TestDateTime, "2004-12-05T12")
        # invalid partial datetime
        with self.assertRaises(ValueError):
            pdatetime = TestText(Name="AETERM", TestDateTime="2004---15T-:05")

    def test_partial_date_type(self):
        # valid partial date
        pdate = TestPartialDate(Name="AETERM", TestDate="2004-12-20")
        self.assertEqual(pdate.TestDate, "2004-12-20")
        pdate = TestPartialDate(Name="AETERM", TestDate="2004-12")
        self.assertEqual(pdate.TestDate, "2004-12")
        pdate = TestPartialDate(Name="AETERM", TestDate="2004")
        self.assertEqual(pdate.TestDate, "2004")
        # invalid partial date
        with self.assertRaises(ValueError):
            pdate = TestPartialDate(Name="AETERM", TestDate="----20")
        with self.assertRaises(ValueError):
            pdate = TestPartialDate(Name="AETERM", TestDate="2004-13-20")

    def test_partial_time_type(self):
        # valid partial time
        ptime = TestPartialDate(Name="AETERM", TestTime="01:01:01Z")
        self.assertEqual(ptime.TestTime, "01:01:01Z")
        ptime = TestPartialDate(Name="AETERM", TestTime="12:05")
        self.assertEqual(ptime.TestTime, "12:05")
        ptime = TestPartialDate(Name="AETERM", TestTime="12")
        self.assertEqual(ptime.TestTime, "12")
        # invalid partial datetime
        with self.assertRaises(ValueError):
            ptime = TestPartialDate(Name="AETERM", TestTime="-:05")
